from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal, cast, overload

from cognite.neat.rules.models._base import DataModelType, ExtensionCategory, SchemaCompleteness, _add_alias
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSUnknownEntity,
    Unknown,
    ViewEntity,
    ViewPropertyEntity,
)

from ._rules import DMSContainer, DMSMetadata, DMSProperty, DMSRules, DMSView


@dataclass
class DMSMetadataInput:
    schema_: Literal["complete", "partial", "extended"]
    space: str
    external_id: str
    creator: str
    version: str
    extension: Literal["addition", "reshape", "rebuild"] = "addition"
    data_model_type: Literal["solution", "enterprise"] = "solution"
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None

    @classmethod
    def load(cls, data: dict[str, Any] | None) -> "DMSMetadataInput | None":
        if data is None:
            return None
        _add_alias(data, DMSMetadata)
        return cls(
            schema_=data.get("schema_"),  # type: ignore[arg-type]
            space=data.get("space"),  # type: ignore[arg-type]
            external_id=data.get("external_id"),  # type: ignore[arg-type]
            creator=data.get("creator"),  # type: ignore[arg-type]
            version=data.get("version"),  # type: ignore[arg-type]
            # safeguard from empty cell, i.e. if key provided by value None
            extension=data.get("extension", "addition") or "addition",
            data_model_type=data.get("data_model_type", "solution") or "solution",
            name=data.get("name"),
            description=data.get("description"),
            created=data.get("created"),
            updated=data.get("updated"),
        )

    def dump(self) -> dict[str, Any]:
        return dict(
            schema=SchemaCompleteness(self.schema_),
            extension=ExtensionCategory(self.extension),
            space=self.space,
            externalId=self.external_id,
            dataModelType=DataModelType(self.data_model_type),
            creator=self.creator,
            version=self.version,
            name=self.name,
            description=self.description,
            created=self.created or datetime.now(),
            updated=self.updated or datetime.now(),
        )


@dataclass
class DMSPropertyInput:
    view: str
    view_property: str | None
    value_type: str
    property_: str | None
    class_: str | None = None
    name: str | None = None
    description: str | None = None
    connection: Literal["direct", "edge", "reverse"] | None = None
    nullable: bool | None = None
    is_list: bool | None = None
    default: str | int | dict | None = None
    reference: str | None = None
    container: str | None = None
    container_property: str | None = None
    index: str | list[str] | None = None
    constraint: str | list[str] | None = None

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "DMSPropertyInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["DMSPropertyInput"]: ...

    @classmethod
    def load(
        cls, data: dict[str, Any] | list[dict[str, Any]] | None
    ) -> "DMSPropertyInput | list[DMSPropertyInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]

        _add_alias(data, DMSProperty)
        return cls(
            view=data.get("view"),  # type: ignore[arg-type]
            view_property=data.get("view_property"),  # type: ignore[arg-type]
            value_type=data.get("value_type"),  # type: ignore[arg-type]
            property_=data.get("property_"),
            class_=data.get("class_"),
            name=data.get("name"),
            description=data.get("description"),
            connection=data.get("connection"),
            nullable=data.get("nullable"),
            is_list=data.get("is_list"),
            default=data.get("default"),
            reference=data.get("reference"),
            container=data.get("container"),
            container_property=data.get("container_property"),
            index=data.get("index"),
            constraint=data.get("constraint"),
        )

    def dump(self, default_space: str, default_version: str) -> dict[str, Any]:
        value_type: DataType | ViewPropertyEntity | ViewEntity | DMSUnknownEntity
        if DataType.is_data_type(self.value_type):
            value_type = DataType.load(self.value_type)
        elif self.value_type == str(Unknown):
            value_type = DMSUnknownEntity()
        else:
            try:
                value_type = ViewPropertyEntity.load(self.value_type, space=default_space, version=default_version)
            except ValueError:
                value_type = ViewEntity.load(self.value_type, space=default_space, version=default_version)

        return {
            "View": ViewEntity.load(self.view, space=default_space, version=default_version),
            "View Property": self.view_property,
            "Value Type": value_type,
            "Property (linage)": self.property_ or self.view_property,
            "Class (linage)": (
                ClassEntity.load(self.class_, prefix=default_space, version=default_version) if self.class_ else None
            ),
            "Name": self.name,
            "Description": self.description,
            "Connection": self.connection,
            "Nullable": self.nullable,
            "Is List": self.is_list,
            "Default": self.default,
            "Reference": self.reference,
            "Container": (
                ContainerEntity.load(self.container, space=default_space, version=default_version)
                if self.container
                else None
            ),
            "Container Property": self.container_property,
            "Index": self.index,
            "Constraint": self.constraint,
        }


@dataclass
class DMSContainerInput:
    container: str
    class_: str | None = None
    name: str | None = None
    description: str | None = None
    reference: str | None = None
    constraint: str | None = None

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "DMSContainerInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["DMSContainerInput"]: ...

    @classmethod
    def load(
        cls, data: dict[str, Any] | list[dict[str, Any]] | None
    ) -> "DMSContainerInput | list[DMSContainerInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]

        _add_alias(data, DMSContainer)
        return cls(
            container=data.get("container"),  # type: ignore[arg-type]
            class_=data.get("class_"),
            name=data.get("name"),
            description=data.get("description"),
            reference=data.get("reference"),
            constraint=data.get("constraint"),
        )

    def dump(self, default_space: str) -> dict[str, Any]:
        container = ContainerEntity.load(self.container, space=default_space)
        return {
            "Container": container,
            "Class (linage)": (
                ClassEntity.load(self.class_, prefix=default_space) if self.class_ else container.as_class()
            ),
            "Name": self.name,
            "Description": self.description,
            "Reference": self.reference,
            "Constraint": (
                [
                    ContainerEntity.load(constraint.strip(), space=default_space)
                    for constraint in self.constraint.split(",")
                ]
                if self.constraint
                else None
            ),
        }


@dataclass
class DMSViewInput:
    view: str
    class_: str | None = None
    name: str | None = None
    description: str | None = None
    implements: str | None = None
    reference: str | None = None
    filter_: Literal["hasData", "nodeType", "rawFilter"] | None = None
    in_model: bool = True

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "DMSViewInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["DMSViewInput"]: ...

    @classmethod
    def load(cls, data: dict[str, Any] | list[dict[str, Any]] | None) -> "DMSViewInput | list[DMSViewInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]
        _add_alias(data, DMSView)

        return cls(
            view=data.get("view"),  # type: ignore[arg-type]
            class_=data.get("class"),
            name=data.get("name"),
            description=data.get("description"),
            implements=data.get("implements"),
            reference=data.get("reference"),
            filter_=data.get("filter_"),
            in_model=data.get("in_model", True),
        )

    def dump(self, default_space: str, default_version: str) -> dict[str, Any]:
        view = ViewEntity.load(self.view, space=default_space, version=default_version)
        return {
            "View": view,
            "Class (linage)": (
                ClassEntity.load(self.class_, prefix=default_space, version=default_version)
                if self.class_
                else view.as_class()
            ),
            "Name": self.name,
            "Description": self.description,
            "Implements": (
                [
                    ViewEntity.load(implement, space=default_space, version=default_version)
                    for implement in self.implements.split(",")
                ]
                if self.implements
                else None
            ),
            "Reference": self.reference,
            "Filter": self.filter_,
            "In Model": self.in_model,
        }


@dataclass
class DMSRulesInput:
    metadata: DMSMetadataInput
    properties: Sequence[DMSPropertyInput]
    views: Sequence[DMSViewInput]
    containers: Sequence[DMSContainerInput] | None = None
    last: "DMSRulesInput | DMSRules | None" = None
    reference: "DMSRulesInput | DMSRules | None" = None

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "DMSRulesInput": ...

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    def load(cls, data: dict | None) -> "DMSRulesInput | None":
        if data is None:
            return None
        _add_alias(data, DMSRules)
        return cls(
            metadata=DMSMetadataInput.load(data.get("metadata")),  # type: ignore[arg-type]
            properties=DMSPropertyInput.load(data.get("properties")),  # type: ignore[arg-type]
            views=DMSViewInput.load(data.get("views")),  # type: ignore[arg-type]
            containers=DMSContainerInput.load(data.get("containers")) or [],
            last=DMSRulesInput.load(data.get("last")),
            reference=DMSRulesInput.load(data.get("reference")),
        )

    def as_rules(self) -> DMSRules:
        return DMSRules.model_validate(self.dump())

    def dump(self) -> dict[str, Any]:
        default_space = self.metadata.space
        default_version = self.metadata.version
        reference: dict[str, Any] | None = None
        if isinstance(self.reference, DMSRulesInput):
            reference = self.reference.dump()
        elif isinstance(self.reference, DMSRules):
            # We need to load through the DMSRulesInput to set the correct default space and version
            reference = DMSRulesInput.load(self.reference.model_dump()).dump()
        last: dict[str, Any] | None = None
        if isinstance(self.last, DMSRulesInput):
            last = self.last.dump()
        elif isinstance(self.last, DMSRules):
            # We need to load through the DMSRulesInput to set the correct default space and version
            last = DMSRulesInput.load(self.last.model_dump()).dump()

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_space, default_version) for prop in self.properties],
            Views=[view.dump(default_space, default_version) for view in self.views],
            Containers=[container.dump(default_space) for container in self.containers or []] or None,
            Last=last,
            Reference=reference,
        )



import xml.etree.ElementTree as ET
from pathlib import Path

from rdflib import Namespace

from cognite.neat.constants import PREFIXES
from cognite.neat.graph.extractors._dexpi import DexpiExtractor
from cognite.neat.legacy.graph.models import Triple

from ._base import BaseExtractor

_DEXPI_PREFIXES = {
    "dexpi": Namespace("http://sandbox.dexpi.org/rdl/"),
    "posccaesar": Namespace("http://data.posccaesar.org/rdl/"),
}


class DexpiXML(BaseExtractor):
    """
    DEXPI-XML extractor of RDF triples

    Args:
        filepath: File path to DEXPI XML file.
        namespace: Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to http://purl.org/cognite/neat/.
    """

    def __init__(
        self,
        filepath: Path | str,
        base_namespace: str | None = None,
    ):
        self.filepath = Path(filepath)
        self.namespace = Namespace(base_namespace) if isinstance(base_namespace, str | Namespace) else PREFIXES["neat"]

    def extract(self) -> set[Triple]:
        """
        Extracts RDF triples from the graph capturing sheet.

        Returns:
            List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
        """
        if self.filepath is None:
            raise ValueError("File path to the graph capturing sheet is not provided!")

        root = ET.parse(self.filepath).getroot()

        # removing legacy code by reusing the maintained version of DexpiExtractor
        return set(DexpiExtractor(root, self.namespace).extract())



from typing import cast

from rdflib import Graph, Literal, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph import extractors

from ._base import BaseTransformer


class AddAssetDepth(BaseTransformer):
    description: str = "Adds depth of asset in the asset hierarchy to the graph"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__)})

    _parent_template: str = """SELECT ?child ?parent WHERE {{
                              <{asset_id}> <{parent_prop}> ?child .
                              OPTIONAL{{?child <{parent_prop}>+ ?parent .}}}}"""

    _root_template: str = """SELECT ?root WHERE {{
                             <{asset_id}> <{root_prop}> ?root .}}"""

    def __init__(
        self, asset_type: URIRef | None = None, root_prop: URIRef | None = None, parent_prop: URIRef | None = None
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.root_prop = root_prop or DEFAULT_NAMESPACE.root
        self.parent_prop = parent_prop or DEFAULT_NAMESPACE.parent

    def transform(self, graph: Graph) -> None:
        """Adds depth of asset in the asset hierarchy to the graph."""
        for result in graph.query(f"SELECT DISTINCT ?asset_id WHERE {{?asset_id a <{self.asset_type}>}}"):
            asset_id = cast(tuple, result)[0]
            if depth := self.get_depth(graph, asset_id, self.root_prop, self.parent_prop):
                graph.add((asset_id, DEFAULT_NAMESPACE.depth, Literal(depth)))

    @classmethod
    def get_depth(
        cls,
        graph: Graph,
        asset_id: URIRef,
        root_prop: URIRef,
        parent_prop: URIRef,
    ) -> int | None:
        """Get asset depth in the asset hierarchy."""

        # Handles non-root assets
        if result := list(graph.query(cls._parent_template.format(asset_id=asset_id, parent_prop=parent_prop))):
            return len(cast(list[tuple], result)) + 2 if cast(list[tuple], result)[0][1] else 2

        # Handles root assets
        elif (
            (result := list(graph.query(cls._root_template.format(asset_id=asset_id, root_prop=root_prop))))
            and len(cast(list[tuple], result)) == 1
            and cast(list[tuple], result)[0][0] == asset_id
        ):
            return 1
        else:
            return None


class AssetTimeSeriesConnector(BaseTransformer):
    description: str = "Connects assets to timeseries, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__), str(extractors.TimeSeriesExtractor.__name__)})
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{timeseries_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        timeseries_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.timeseries_type = timeseries_type or DEFAULT_NAMESPACE.TimeSeries
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for ts_id_result in graph.query(
            f"SELECT DISTINCT ?timeseries_id WHERE {{?timeseries_id a <{self.timeseries_type}>}}"
        ):
            timeseries_id: URIRef = cast(tuple, ts_id_result)[0]

            if asset_id_res := list(
                graph.query(
                    self._asset_template.format(
                        timeseries_id=timeseries_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # timeseries can be connected to only one asset in the graph
                asset_id = cast(list[tuple], asset_id_res)[0][0]
                graph.add((asset_id, DEFAULT_NAMESPACE.timeSeries, timeseries_id))


class AssetSequenceConnector(BaseTransformer):
    description: str = "Connects assets to sequences, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__), str(extractors.SequencesExtractor.__name__)})
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{sequence_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        sequence_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.sequence_type = sequence_type or DEFAULT_NAMESPACE.Sequence
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for sequency_id_result in graph.query(
            f"SELECT DISTINCT ?sequence_id WHERE {{?sequence_id a <{self.sequence_type}>}}"
        ):
            sequence_id: URIRef = cast(tuple, sequency_id_result)[0]

            if asset_id_res := list(
                graph.query(
                    self._asset_template.format(
                        sequence_id=sequence_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # sequence can be connected to only one asset in the graph
                asset_id = cast(list[tuple], asset_id_res)[0][0]
                graph.add((asset_id, DEFAULT_NAMESPACE.sequence, sequence_id))


class AssetFileConnector(BaseTransformer):
    description: str = "Connects assets to files, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__), str(extractors.FilesExtractor.__name__)})
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{file_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        file_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.file_type = file_type or DEFAULT_NAMESPACE.File
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for sequency_id_result in graph.query(f"SELECT DISTINCT ?file_id WHERE {{?file_id a <{self.file_type}>}}"):
            file_id: URIRef = cast(tuple, sequency_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        file_id=file_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for (asset_id,) in cast(list[tuple], assets_id_res):
                    graph.add((asset_id, DEFAULT_NAMESPACE.file, file_id))


class AssetEventConnector(BaseTransformer):
    description: str = "Connects assets to events, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__), str(extractors.EventsExtractor.__name__)})
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{event_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        event_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.event_type = event_type or DEFAULT_NAMESPACE.Event
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for event_id_result in graph.query(f"SELECT DISTINCT ?event_id WHERE {{?event_id a <{self.event_type}>}}"):
            event_id: URIRef = cast(tuple, event_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        event_id=event_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for (asset_id,) in cast(list[tuple], assets_id_res):
                    graph.add((asset_id, DEFAULT_NAMESPACE.event, event_id))


class AssetRelationshipConnector(BaseTransformer):
    description: str = "Connects assets via relationships"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {str(extractors.AssetsExtractor.__name__), str(extractors.RelationshipsExtractor.__name__)}
    )
    _asset_template: str = """SELECT ?source ?target WHERE {{
                              <{relationship_id}> <{relationship_source_xid_prop}> ?source_xid .
                              ?source <{asset_xid_property}> ?source_xid .
                              ?source a <{asset_type}> .

                              <{relationship_id}> <{relationship_target_xid_prop}> ?target_xid .
                              ?target <{asset_xid_property}> ?target_xid .
                              ?target a <{asset_type}> .}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        relationship_type: URIRef | None = None,
        relationship_source_xid_prop: URIRef | None = None,
        relationship_target_xid_prop: URIRef | None = None,
        asset_xid_property: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.relationship_type = relationship_type or DEFAULT_NAMESPACE.Relationship
        self.relationship_source_xid_prop = relationship_source_xid_prop or DEFAULT_NAMESPACE.source_external_id
        self.relationship_target_xid_prop = relationship_target_xid_prop or DEFAULT_NAMESPACE.target_external_id
        self.asset_xid_property = asset_xid_property or DEFAULT_NAMESPACE.external_id

    def transform(self, graph: Graph) -> None:
        for relationship_id_result in graph.query(
            f"SELECT DISTINCT ?relationship_id WHERE {{?relationship_id a <{self.relationship_type}>}}"
        ):
            relationship_id: URIRef = cast(tuple, relationship_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        relationship_id=relationship_id,
                        asset_xid_property=self.asset_xid_property,
                        relationship_source_xid_prop=self.relationship_source_xid_prop,
                        relationship_target_xid_prop=self.relationship_target_xid_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for source_asset_id, target_asset_id in cast(list[tuple], assets_id_res):
                    # create a relationship between the two assets
                    graph.add((source_asset_id, DEFAULT_NAMESPACE.relationship, relationship_id))
                    graph.add((target_asset_id, DEFAULT_NAMESPACE.relationship, relationship_id))

                    # add source and target to the relationship
                    graph.add((relationship_id, DEFAULT_NAMESPACE.source, source_asset_id))
                    graph.add((relationship_id, DEFAULT_NAMESPACE.target, target_asset_id))

                    # remove properties that are not needed, specifically the external ids
                    graph.remove((relationship_id, self.relationship_source_xid_prop, None))
                    graph.remove((relationship_id, self.relationship_target_xid_prop, None))



from ._base import BaseLoader, CDFLoader
from ._rdf2dms import DMSLoader

__all__ = ["BaseLoader", "CDFLoader", "DMSLoader"]


def _repr_html_() -> str:
    import pandas as pd

    table = pd.DataFrame(  # type: ignore[operator]
        [
            {
                "Loader": name,
                "Description": globals()[name].__doc__.strip().split("\n")[0] if globals()[name].__doc__ else "Missing",
            }
            for name in __all__
            if name not in ("BaseLoader", "CDFLoader")
        ]
    )._repr_html_()

    return (
        "<strong>Loader</strong> A loader writes data from Neat's triple storage into a target system" f"<br />{table}"
    )



from cognite.neat.legacy.rules.models.rules import Rules


def get_labels(transformation_rules: Rules) -> set[str]:
    """
    Return CDF labels for classes and relationships.

    Args:
        transformation_rules: The transformation rules to extract labels from.

    Returns:
        Set of CDF labels
    """
    class_labels = {class_.class_id for class_ in transformation_rules.classes.values()}

    property_labels = {property_.property_id for property_ in transformation_rules.properties.values()}

    relationship_labels = {
        str(rule.label)
        for rule in transformation_rules.properties.values()
        if "Relationship" in rule.cdf_resource_type and rule.label
    }

    return class_labels.union(relationship_labels).union(property_labels)



from typing import ClassVar

from cognite.client import CogniteClient

from cognite.neat.legacy.graph.transformations.transformer import RuleProcessingReport, domain2app_knowledge_graph
from cognite.neat.legacy.rules.exporters._rules2triples import get_instances_as_triples
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import RulesData, SolutionGraph, SourceGraph
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = ["TransformSourceToSolutionGraph"]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title() + " [LEGACY]"


class TransformSourceToSolutionGraph(Step):
    """
    The step transforms source graph to solution graph
    """

    description = "The step transforms source graph to solution graph"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="cdf_lookup_database",
            value="",
            label="Name of the CDF raw database to use for data lookup (rawlookup rules).\
            Applicable only for transformations with rawlookup rules.",
        )
    ]

    def run(  # type: ignore[override, syntax]
        self,
        transformation_rules: RulesData,
        cdf_client: CogniteClient,
        source_graph: SourceGraph,
        solution_graph: SolutionGraph,
    ) -> FlowMessage:
        report = RuleProcessingReport()
        # run transformation and generate new graph
        cdf_lookup_database = self.configs.get("cdf_lookup_database", "")
        solution_graph.graph.set_graph(
            domain2app_knowledge_graph(
                source_graph.graph.get_graph(),
                transformation_rules.rules,
                app_instance_graph=solution_graph.graph.get_graph(),
                extra_triples=get_instances_as_triples(transformation_rules.rules),
                client=cdf_client,
                cdf_lookup_database=cdf_lookup_database,  # change this accordingly!
                processing_report=report,
            )
        )
        return FlowMessage(
            output_text=f"Total processed rules: { report.total_rules } , success: { report.total_success } , \
             no results: { report.total_success_no_results } , failed: { report.total_failed }",
            payload=report,
        )



import importlib
import inspect
from collections.abc import Callable
from types import ModuleType

from cognite.neat.exceptions import NeatImportError


def local_import(module: str, extra: str) -> ModuleType:
    try:
        return importlib.import_module(module)
    except ImportError as e:
        raise NeatImportError(module.split(".")[0], extra) from e


def get_classmethods(cls: type) -> list[Callable]:
    return [
        func for _, func in inspect.getmembers(cls, lambda x: inspect.ismethod(x) and not x.__name__.startswith("_"))
    ]


def class_html_doc(cls: type, include_factory_methods: bool = True) -> str:
    if cls.__doc__:
        docstring = cls.__doc__.split("Args:")[0].strip().replace("\n", "<br />")
    else:
        docstring = "Missing Description"
    if include_factory_methods:
        factory_methods = get_classmethods(cls)
        if factory_methods:
            factory_methods_str = "".join(f"<li><em>.{m.__name__}</em></li>" for m in factory_methods)
            docstring += (
                f"<br /><strong>Available factory methods:</strong><br />"
                f'<ul style="list-style-type:circle;">{factory_methods_str}</ul>'
            )
    return f"<h3>{cls.__name__}</h3><p>{docstring}</p>"



from typing import Any
from warnings import WarningMessage

from pydantic_core import ErrorDetails, PydanticCustomError


class NeatException(Exception):
    """Base class for all exceptions raised by NEAT."""

    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str

    def to_pydantic_custom_error(self):
        return PydanticCustomError(
            self.type_,
            self.message,
            dict(type_=self.type_, code=self.code, description=self.description, example=self.example, fix=self.fix),
        )

    def to_error_dict(self) -> ErrorDetails:
        return {
            "type": self.type_,
            "loc": (),
            "msg": self.message,
            "input": None,
            "ctx": dict(
                type_=self.type_, code=self.code, description=self.description, example=self.example, fix=self.fix
            ),
        }


class NeatWarning(UserWarning):
    type_: str
    code: int
    description: str
    example: str
    fix: str
    message: str


class NeatImportError(NeatException):
    """Cognite Import Error

    Raised if the user attempts to use functionality which requires an uninstalled package.

    Args:
        module (str): Name of the module which could not be imported
        extra (str): The name of the extra you use to install it with neat.
    """

    type_: str = "NeatImportError"
    code: int = 0
    description: str = "Raised if the user attempts to use functionality which requires an uninstalled package."
    example: str = ""
    fix: str = ""

    def __init__(self, module: str, extra: str, verbose=False):
        self.module = module
        self.message = (
            f"This functionality requires {self.module}. "
            f'You can include it in your neat installation with `pip install "cognite-neat[{extra}]"`.'
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class InvalidWorkFlowError(NeatException):
    """InvalidWorkFlowError
    Raised if an invalid workflow is provided to the Workflow Manager.
    Args:
        step_name (str): Name of the step which could not be run
        missing_data (list[str]): The missing data for the step.
    """

    type_: str = "InvalidWorkFlowError"
    code: int = 1
    description: str = "Raised if an invalid workflow is provided to the Workflow Manager."
    example: str = ""
    fix: str = ""

    def __init__(self, step_name, missing_data: list[str], verbose=False):
        self.message = f"In the workflow step {step_name} the following data is missing: {missing_data}."

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


def wrangle_warnings(list_of_warnings: list[WarningMessage]) -> list[dict]:
    warning_list: list[dict] = []
    for warning in list_of_warnings:
        if issubclass(warning.message.__class__, NeatWarning):
            warning_list.append(_neat_warning_to_dict(warning))
        elif issubclass(warning.message.__class__, Warning):
            warning_list.append(_python_warning_to_dict(warning))
    return warning_list


def _neat_warning_to_dict(warning: WarningMessage) -> dict:
    category: Any = warning.category
    return {
        "type": category.type_,
        "loc": (),
        "msg": str(warning.message),
        "input": None,
        "ctx": dict(
            type_=category.type_,
            code=category.code,
            description=category.description,
            example=category.example,
            fix=category.fix,
        ),
    }


def _python_warning_to_dict(warning: WarningMessage) -> dict:
    return {
        "type": warning.category,
        "loc": (),
        "msg": str(warning.message),
        "input": None,
        "ctx": dict(type_=warning.category, code=None, description=None, example=None, fix=None),
    }



from collections.abc import Generator
from xml.etree.ElementTree import Element


def iterate_tree(node: Element) -> Generator:
    """Iterate over all elements in an XML tree.

    Args:
        node: XML tree to iterate over.

    Returns:
        Generator of XML elements.
    """
    yield node
    for child in node:
        yield from iterate_tree(child)


def get_children(element: Element, child_tag: str, no_children: int = -1) -> list[Element]:
    """Get children of an XML element.

    Args:
        element: XML element to get children from.
        child_tag: Tag of the children to get.
        no_children: Max number of children to get. Defaults to -1 (all).

    Returns:
        List of XML elements if no_children > 1, otherwise XML element.
    """
    children = []
    for child in element:
        if child.tag == child_tag:
            if no_children == 1:
                return [child]
            else:
                children.append(child)
    return children



import hashlib
import json
import logging
import uuid
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import ClassVar, cast

from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat.constants import PREFIXES
from cognite.neat.legacy.graph import extractors
from cognite.neat.legacy.graph.extractors._mock_graph_generator import generate_triples as generate_mock_triples
from cognite.neat.legacy.rules.exporters._rules2triples import get_instances_as_triples
from cognite.neat.utils.utils import create_sha256_hash
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import RulesData, SolutionGraph, SourceGraph
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "ExtractGraphFromRdfFile",
    "ExtractGraphFromRulesInstanceSheet",
    "ExtractGraphFromGraphCapturingSheet",
    "ExtractGraphFromMockGraph",
    "ExtractGraphFromRulesDataModel",
    "ExtractGraphFromJsonFile",
    "ExtractGraphFromAvevaPiAssetFramework",
    "ExtractGraphFromDexpiFile",
]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title() + " [LEGACY]"


class ExtractGraphFromRdfFile(Step):
    """
    This step extract instances from a file into the source graph. The file must be in RDF format.
    """

    description = "This step extract instances from a file into the source graph. The file must be in RDF format."
    version = "legacy"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_path",
            value="source-graphs/source-graph-dump.xml",
            label="File name of source graph data dump in RDF format",
        ),
        Configurable(
            name="mime_type",
            value="application/rdf+xml",
            label="MIME type of file containing RDF graph",
            options=[
                "application/rdf+xml",
                "text/turtle",
                "application/n-triples",
                "application/n-quads",
                "application/trig",
            ],
        ),
        Configurable(
            name="add_base_iri",
            value="True",
            label="Whether to add base IRI to graph in case if entity ids are relative",
            options=["True", "False"],
        ),
    ]

    def run(self, source_graph: SourceGraph) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        if source_graph.graph.rdf_store_type.lower() in ("memory", "oxigraph"):
            if source_file := self.configs["file_path"]:
                source_graph.graph.import_from_file(
                    self.data_store_path / Path(source_file),
                    mime_type=self.configs["mime_type"],  # type: ignore[arg-type]
                    add_base_iri=self.configs["add_base_iri"] == "True",
                )
                logging.info(f"Loaded {source_file} into source graph.")
            else:
                raise ValueError("You need a source_rdf_store.file specified for source_rdf_store.type=memory")
        else:
            raise NotImplementedError(f"Graph type {source_graph.graph.rdf_store_type} is not supported.")

        return FlowMessage(output_text="Instances loaded to source graph")


class ExtractGraphFromDexpiFile(Step):
    """
    This step converts DEXPI P&ID (XML) into Knowledge Graph
    """

    description = "This step converts DEXPI P&ID (XML) into Knowledge Graph"
    version = "legacy"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_path",
            value="source-graphs/dexpi-pid.xml",
            label="File path to DEXPI P&ID in XML format",
        ),
        Configurable(
            name="base_namespace",
            value="http://purl.org/cognite/neat#",
            label="Base namespace to be added to ids for all nodes found in P&ID",
        ),
    ]

    def run(self, source_graph: SourceGraph) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        file_path = self.configs.get("file_path")
        base_namespace = self.configs.get("base_namespace", None)

        if file_path:
            triples = extractors.DexpiXML(self.data_store_path / Path(file_path), base_namespace).extract()
            source_graph.graph.add_triples(triples, verbose=True)

            logging.info(f"Loaded {file_path} into source graph.")
        else:
            raise ValueError("You need a source_rdf_store.file specified")

        return FlowMessage(output_text="Instances loaded to source graph")


class ExtractGraphFromGraphCapturingSheet(Step):
    """
    This step extracts nodes and edges from graph capture spreadsheet and load them into graph
    """

    description = "This step extracts nodes and edges from graph capturing spreadsheet and load them into graph"
    version = "legacy"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_path",
            value="source-graphs/graph_capture_sheet.xlsx",
            label="File path to Graph Capturing Sheet",
        ),
        Configurable(
            name="base_namespace",
            value="http://purl.org/cognite/neat#",
            label="Base namespace to be added to ids for all nodes extracted from graph capturing spreadsheet",
        ),
        Configurable(
            name="graph_name",
            value="solution",
            label="The name of target graph to load nodes and edge sto.",
            options=["source", "solution"],
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, rules: RulesData, graph_store: SolutionGraph | SourceGraph
    ) -> FlowMessage:
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        file_path = self.configs.get("file_path")

        if file_path:
            logging.info(f"Processing graph capture sheet {self.data_store_path / Path(file_path)}")

            triples = extractors.GraphCapturingSheet(
                rules=rules.rules,
                filepath=self.data_store_path / Path(file_path),
                namespace=self.configs.get("base_namespace", None),
                use_source_ids=True,
            ).extract()

        else:
            raise ValueError("You need a source_rdf_store.file specified")

        if self.configs["graph_name"] == "solution":
            graph_store = cast(SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph_store = cast(SourceGraph, self.flow_context["SourceGraph"])

        graph_store.graph.add_triples(triples, verbose=True)  # type: ignore[arg-type]
        return FlowMessage(output_text="Graph capture sheet processed")


class ExtractGraphFromMockGraph(Step):
    """
    This step generate mock graph based on the defined classes and target number of instances
    """

    description = "This step generate mock graph based on the defined classes and target number of instances"
    version = "legacy"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="class_count",
            value='{"GeographicalRegion":5, "SubGeographicalRegion":10}',
            label="Target number of instances for each class",
        ),
        Configurable(
            name="graph_name", value="solution", label="The name of target graph.", options=["source", "solution"]
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, transformation_rules: RulesData, graph_store: SolutionGraph | SourceGraph
    ) -> FlowMessage:
        if self.configs is None:
            raise StepNotInitialized(type(self).__name__)
        logging.info("Initiated generation of mock triples")
        try:
            class_count = json.loads(self.configs["class_count"])
        except Exception:
            return FlowMessage(
                error_text="Defected JSON stored in class_count",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        if self.configs["graph_name"] == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            graph_store = cast(SourceGraph | SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph_store = cast(SourceGraph | SolutionGraph, self.flow_context["SourceGraph"])

        logging.info(class_count)
        logging.info(transformation_rules.rules.metadata.model_dump())
        try:
            triples = generate_mock_triples(transformation_rules=transformation_rules.rules, class_count=class_count)
        except Exception as e:
            return FlowMessage(error_text=f"Error: {e}", step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        logging.info("Adding mock triples to graph")
        graph_store.graph.add_triples(triples, verbose=True)  # type: ignore[arg-type]
        return FlowMessage(output_text=f"Mock graph generated containing total of {len(triples)} triples")


class ExtractGraphFromRulesInstanceSheet(Step):
    """
    This step extracts instances from Rules object and loads them into the graph
    """

    description = "This step extracts instances from Rules object and loads them into the graph."
    category = CATEGORY
    version = "legacy"

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="graph_name", value="solution", label="The name of target graph.", options=["source", "solution"]
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, transformation_rules: RulesData, graph_store: SolutionGraph | SourceGraph
    ) -> FlowMessage:
        triples = get_instances_as_triples(transformation_rules.rules)
        instance_ids = {triple[0] for triple in triples}
        output_text = f"Extracted {len(instance_ids)} instances out of"
        output_text += f"Loaded {len(triples)} statements defining"
        output_text += f" {len(instance_ids)} instances"

        if self.configs["graph_name"] == "solution":
            graph_store = cast(SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph_store = cast(SourceGraph, self.flow_context["SourceGraph"])

        try:
            graph_store.graph.add_triples(triples, verbose=True)  # type: ignore[arg-type]
        except Exception as e:
            return FlowMessage(error_text=f"Error: {e}", step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        return FlowMessage(output_text=output_text)


class ExtractGraphFromRulesDataModel(Step):
    """
    This step extracts data model from rules file and loads it into source graph
    """

    description = "This step extracts data model from rules file and loads it into source graph."
    category = CATEGORY
    version = "legacy"

    def run(  # type: ignore[override, syntax]
        self, transformation_rules: RulesData, source_graph: SourceGraph
    ) -> FlowMessage:
        ns = PREFIXES["neat"]
        classes = transformation_rules.rules.classes
        properties = transformation_rules.rules.properties
        counter = 0
        for class_name, class_def in classes.items():
            rdf_instance_id = URIRef(ns + "_" + class_def.class_id)
            source_graph.graph.graph.add((rdf_instance_id, URIRef(ns + "Name"), Literal(class_name)))
            source_graph.graph.graph.add((rdf_instance_id, RDF.type, URIRef(ns + class_def.class_id)))
            if class_def.parent_class:
                source_graph.graph.graph.add(
                    (rdf_instance_id, URIRef(ns + "hasParent"), URIRef(ns + "_" + cast(str, class_def.parent_class)))
                )
            counter += 1

        for _property_name, property_def in properties.items():
            rdf_instance_id = URIRef(ns + "_" + property_def.class_id)
            source_graph.graph.graph.add(
                (rdf_instance_id, URIRef(ns + property_def.property_id), Literal(property_def.expected_value_type))
            )
            if property_def.expected_value_type.suffix not in ("string", "integer", "float", "boolean"):
                source_graph.graph.graph.add(
                    (
                        rdf_instance_id,
                        URIRef(ns + "connectedTo"),
                        URIRef(ns + "_" + property_def.expected_value_type.suffix),
                    )
                )
            counter += 1

        output_text = f"Loaded {counter} classes into source graph"
        return FlowMessage(output_text=output_text)


class ExtractGraphFromJsonFile(Step):
    """
    This step extracts instances from json file and loads them into a graph store. Warning : the step is experimental
    """

    description = "This step extracts instances from json file and loads them into a graph store"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name", value="data_dump.json", label="Full path to the file containing data dump in JSON format"
        ),
        Configurable(
            name="graph_name", value="solution", label="The name of target graph.", options=["source", "solution"]
        ),
        Configurable(
            name="object_id_generation_method",
            value="hash_of_json_element",
            label="Method to be used for generating object ids.  \
                 source_object_properties - takes multiple properties from the source object and concatenates them. \
                 source_object_id_mapping - takes a single property from the \
                 source object and maps it to a instance id. \
                      The option should be used when source object already contains stable ids \
                hash_of_json_element - takes a hash of the JSON element.Very generic method but \
                     can be slow working with big objects. \
                uuid - generates a random UUID, the option produces unstables ids . ",
            options=["source_object_properties", "source_object_id_mapping", "hash_of_json_element", "uuid"],
        ),
        Configurable(
            name="json_object_id_mapping",
            value="name",
            label="Comma separated list of object properties to be used for generating object ids. \
            Each property must be prefixed with the name of the object. For example: device:name,pump:id",
        ),
        Configurable(
            name="json_object_labels_mapping",
            value="",
            label="Comma separated list of object properties to be used for generating object labels. \
            Each property must be prefixed with the name of the object. For example: asset:name,asset:type",
        ),
        Configurable(
            name="namespace",
            value="http://purl.org/cognite/neat#",
            label="Namespace to be used for the generated objects.",
        ),
        Configurable(name="namespace_prefix", value="neat", label="The prefix to be used for the namespace."),
    ]

    def get_json_object_id(self, method, object_name: str, json_object: dict, parent_object_id: str, id_mapping: dict):
        if method == "source_object_properties":
            object_id = ""
            if object_name in id_mapping:
                for property_name in id_mapping[object_name]:
                    object_id += property_name + json_object[property_name]
        elif method == "hash_of_json_element":
            flat_json_object = {}
            for key, value in json_object.items():
                if not isinstance(value, dict) and not isinstance(value, list):
                    flat_json_object[key] = value

            object_id = json.dumps(flat_json_object, sort_keys=True)
        elif method == "uuid":
            return uuid.uuid4()
        elif method == "source_object_id_mapping":
            # don't hash existing valid ids
            try:
                return json_object[id_mapping[object_name][0]]
            except KeyError as e:
                # back to hashing
                logging.debug(f"Object {object_name} doesn't have a valid id.Error : {e}")
                object_id = self.get_json_object_id(
                    "hash_of_json_element", object_name, json_object, parent_object_id, id_mapping
                )
        else:
            raise ValueError(
                f"Unknown object_id_generation_method: {(self.configs or {}).get('object_id_generation_method')}"
            )

        return hashlib.sha256(object_id.encode()).hexdigest()

    def run(self, graph_store: SolutionGraph | SourceGraph) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        # self.graph.bind
        if self.configs["graph_name"] == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            graph_store = cast(SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph_store = cast(SourceGraph, self.flow_context["SourceGraph"])

        ns = Namespace(self.configs["namespace"])
        graph_store.graph.graph.bind(self.configs["namespace_prefix"], ns)

        full_path = self.data_store_path / Path(self.configs["file_name"])
        logging.info(f"Loading data dump from {full_path}")
        with full_path.open() as f:
            json_data = json.load(f)

        graph = graph_store.graph
        nodes_counter = 0
        property_counter = 0
        labels_mapping: dict[str, str] = {}
        object_id_mapping: dict[str, list[str]] = {}
        if self.configs["json_object_labels_mapping"]:
            for label_mapping in self.configs["json_object_labels_mapping"].split(","):
                object_name, property_name = label_mapping.split(":")
                labels_mapping[object_name] = property_name

        if self.configs["json_object_id_mapping"]:
            for id_mapping in self.configs["json_object_id_mapping"].split(","):
                if ":" not in id_mapping:
                    continue
                object_name, property_name = id_mapping.split(":")
                # if multiple ids are used for the same object ,the order of the properties is important
                if object_name in object_id_mapping:
                    object_id_mapping[object_name].append(property_name)
                else:
                    object_id_mapping[object_name] = [property_name]

        # Iterate through the JSON data and convert it to triples
        def convert_json_to_triples(
            data: dict, parent_node: URIRef, parent_object_id: str, parent_node_path: str, property_name=None
        ):
            nonlocal nodes_counter, property_counter
            if isinstance(data, dict):
                if len(data) == 0:
                    return
                if property_name is None:
                    for key, value in data.items():
                        convert_json_to_triples(value, parent_node, parent_object_id, parent_node_path, key)
                else:
                    object_id = self.get_json_object_id(
                        self.configs["object_id_generation_method"],
                        property_name,
                        data,
                        parent_object_id,
                        object_id_mapping,
                    )
                    new_node = URIRef(ns + object_id)
                    graph.graph.add((new_node, RDF.type, URIRef(ns + property_name)))
                    if labels_mapping and property_name in labels_mapping:
                        graph.graph.add((new_node, URIRef(ns + "label"), Literal(data[labels_mapping[property_name]])))
                    else:
                        graph.graph.add((new_node, URIRef(ns + "label"), Literal(property_name)))
                    graph.graph.add((new_node, URIRef(ns + "parent"), parent_node))
                    nodes_counter += 1
                    for key, value in data.items():
                        new_node_path = parent_node_path + "/" + key
                        convert_json_to_triples(value, new_node, object_id, new_node_path, key)
            elif isinstance(data, list):
                if property_name is None:
                    for key, value in data.items():
                        convert_json_to_triples(value, parent_node, parent_object_id, parent_node_path, key)
                else:
                    for item in data:
                        convert_json_to_triples(item, parent_node, parent_object_id, parent_node_path, property_name)
            else:
                # Convert scalar values to RDF literals
                if isinstance(data, bool):
                    data = Literal(data, datatype=XSD.boolean)
                elif isinstance(data, int):
                    data = Literal(data, datatype=XSD.integer)
                elif isinstance(data, float):
                    data = Literal(data, datatype=XSD.float)
                elif isinstance(data, str):
                    data = Literal(data, datatype=XSD.string)
                else:
                    data = Literal(str(data))
                property_counter += 1
                graph.graph.add((parent_node, URIRef(ns + property_name), data))

        # Start conversion with a root node
        root_node = URIRef(ns + "root")
        graph.graph.add((root_node, URIRef(ns + "label"), Literal("root node")))
        graph.graph.add((root_node, RDF.type, URIRef(ns + "root_node_id")))
        convert_json_to_triples(json_data, root_node, "root", "root", None)
        return FlowMessage(
            output_text=f"Data from source file imported successfully. Imported {nodes_counter} objects \
                            and {property_counter} properties ."
        )


class ExtractGraphFromAvevaPiAssetFramework(Step):
    """
    This step extracts instances from Aveva PI AF and loads them into a graph store. Warning : the step is experimental
    """

    description = "This step extracts instances from Aveva PI AF and loads them into a graph store"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name",
            value="staging/pi_af_dump.xml",
            label="Full path to the file \
            containing data dump in XML format",
        ),
        Configurable(
            name="graph_name", value="solution", label="The name of target graph.", options=["source", "solution"]
        ),
        Configurable(
            name="root_node_external_id",
            value="root",
            label="External id of the root node. The node will be created if it doesn't exist",
        ),
        Configurable(
            name="root_node_name",
            value="root",
            label="Name of the root node. The node will be created if it doesn't exist",
        ),
        Configurable(
            name="root_node_type",
            value="Asset",
            label="Type of the root node. The node will be created if it doesn't exist",
        ),
        Configurable(
            name="namespace",
            value="http://purl.org/cognite/neat#",
            label="Namespace to be used for the generated objects.",
        ),
        Configurable(name="namespace_prefix", value="neat", label="The prefix to be used for the namespace."),
    ]

    def add_root_asset_to_source_graph(self) -> str:
        root_external_id = self.configs["root_node_external_id"]
        root_name = self.configs["root_node_name"]
        root_asset_type = self.configs["root_node_type"]
        rdf_root_instance_id = URIRef(self.ns + root_external_id)
        self.graph_store.graph.add((rdf_root_instance_id, URIRef(self.ns + "Name"), Literal(root_name)))
        self.graph_store.graph.add((rdf_root_instance_id, RDF.type, URIRef(self.ns + root_asset_type)))
        return root_external_id

    def run(  # type: ignore[override, syntax]
        self, flow_msg: FlowMessage, graph_store: SolutionGraph | SourceGraph
    ) -> FlowMessage:
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        if source_file := self.configs["file_name"]:
            source_pi_dump = Path(self.data_store_path) / source_file
        else:
            return FlowMessage(output_text="No source file specified", next_step_ids=["step_error_handler"])

        # self.graph.bind
        if self.configs["graph_name"] == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            self.graph_store = cast(SourceGraph | SolutionGraph, self.flow_context["SolutionGraph"]).graph
        else:
            self.graph_store = cast(SourceGraph | SolutionGraph, self.flow_context["SourceGraph"]).graph

        self.ns = Namespace(self.configs["namespace"])
        self.graph_store.graph.bind(self.configs["namespace_prefix"], self.ns)

        cdf_root_instance_id = self.add_root_asset_to_source_graph()
        # Parse the XML data into an ElementTree object
        root = ET.parse(source_pi_dump).getroot()
        counter = 0
        root_af_element = root.find("AFDatabase/AFElement")
        logging.info(f"Found AFElement: {root_af_element}")

        def process_af_attribute(af_element, element_path=None, parent_element_id: str | None = None):
            name = af_element.find("Name").text
            name = name.replace(" ", "_")
            new_element_path = element_path + "/" + name
            element_id = "_" + create_sha256_hash(new_element_path)
            rdf_instance_id = URIRef(self.ns + element_id)
            self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + "Name"), Literal(name)))
            self.graph_store.graph.add((rdf_instance_id, RDF.type, URIRef(self.ns + "Attribute" + name)))
            self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + "Path"), Literal(new_element_path)))
            if parent_element_id:
                self.graph_store.graph.add(
                    (rdf_instance_id, URIRef(self.ns + "hasParent"), URIRef(self.ns + parent_element_id))
                )
            for child in af_element:
                if child.tag == "AFAttribute":
                    process_af_attribute(child, new_element_path, element_id)
                elif child.tag == "Name":
                    pass
                else:
                    try:
                        self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + child.tag), Literal(child.text)))
                    except Exception as e:
                        logging.error(f"Error parsing AFAttribute {name} : {e}")

        def process_af_element(af_element, element_path=None, parent_element_id: str | None = None) -> str:
            nonlocal counter
            name = af_element.find("Name").text
            template = None
            new_element_path = element_path + "/" + name
            element_id = "_" + create_sha256_hash(new_element_path)
            rdf_instance_id = URIRef(self.ns + element_id)
            self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + "Name"), Literal(name)))
            self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + "Path"), Literal(new_element_path)))
            if parent_element_id:
                self.graph_store.graph.add(
                    (rdf_instance_id, URIRef(self.ns + "hasParent"), URIRef(self.ns + parent_element_id))
                )

            for child in af_element:
                if child.tag == "Name":
                    pass
                if child.tag == "Template":
                    template = child.text
                if child.tag == "AFAttribute":
                    process_af_attribute(child, new_element_path, element_id)
                if child.tag == "AFElement":
                    counter += 1
                    process_af_element(child, new_element_path, element_id)
                else:
                    self.graph_store.graph.add((rdf_instance_id, URIRef(self.ns + child.tag), Literal(child.text)))

            if template:
                self.graph_store.graph.add((rdf_instance_id, RDF.type, URIRef(self.ns + template)))
            else:
                self.graph_store.graph.add((rdf_instance_id, RDF.type, URIRef(self.ns + "AFElement")))

            return element_id

        process_af_element(root_af_element, "root", cdf_root_instance_id)
        self.graph_store.restart()  # restarting the graph to release the memory
        return FlowMessage(output_text=f" {counter} PI assets loaded into the graph")

    def convert_attribute(self, attribute):
        if "{" not in attribute:
            return attribute
        attr_splitted = attribute.split("{")[-1].split("}")
        return attr_splitted[0] + "/" + attr_splitted[1]



import math
import re
import sys
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from cognite.client import data_modeling as dm
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic.main import IncEx
from pydantic_core.core_schema import ValidationInfo

from cognite.neat.rules import issues
from cognite.neat.rules.issues import MultiValueError
from cognite.neat.rules.models._base import (
    BaseMetadata,
    BaseRules,
    DataModelType,
    ExtensionCategory,
    RoleTypes,
    SchemaCompleteness,
    SheetEntity,
    SheetList,
)
from cognite.neat.rules.models._types import (
    ExternalIdType,
    PropertyType,
    StrListType,
    VersionType,
)
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.domain import DomainRules
from cognite.neat.rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    ContainerEntityList,
    DMSUnknownEntity,
    ReferenceEntity,
    URLEntity,
    ViewEntity,
    ViewEntityList,
    ViewPropertyEntity,
)
from cognite.neat.rules.models.wrapped_entities import HasDataFilter, NodeTypeFilter, RawFilter

from ._schema import DMSSchema

if TYPE_CHECKING:
    from cognite.neat.rules.models.information._rules import InformationRules

if sys.version_info >= (3, 11):
    pass
else:
    pass

_DEFAULT_VERSION = "1"


class DMSMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.dms_architect
    data_model_type: DataModelType = Field(DataModelType.enterprise, alias="dataModelType")
    schema_: SchemaCompleteness = Field(alias="schema")
    extension: ExtensionCategory = ExtensionCategory.addition
    space: ExternalIdType
    name: str | None = Field(
        None,
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, min_length=1, max_length=1024)
    external_id: ExternalIdType = Field(alias="externalId")
    version: VersionType
    creator: StrListType
    created: datetime = Field(
        description=("Date of the data model creation"),
    )
    updated: datetime = Field(
        description=("Date of the data model update"),
    )

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value

    @field_serializer("schema_", "extension", "data_model_type", when_used="always")
    @staticmethod
    def as_string(value: SchemaCompleteness | ExtensionCategory | DataModelType) -> str:
        return str(value)

    @field_validator("schema_", mode="plain")
    def as_enum_schema(cls, value: str) -> SchemaCompleteness:
        return SchemaCompleteness(value)

    @field_validator("extension", mode="plain")
    def as_enum_extension(cls, value: str) -> ExtensionCategory:
        return ExtensionCategory(value)

    @field_validator("data_model_type", mode="plain")
    def as_enum_model_type(cls, value: str) -> DataModelType:
        return DataModelType(value)

    @field_validator("description", mode="before")
    def nan_as_none(cls, value):
        if isinstance(value, float) and math.isnan(value):
            return None
        return value

    def as_space(self) -> dm.SpaceApply:
        return dm.SpaceApply(
            space=self.space,
        )

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    def as_data_model(self) -> dm.DataModelApply:
        suffix = f"Creator: {', '.join(self.creator)}"
        if self.description:
            description = f"{self.description} Creator: {', '.join(self.creator)}"
        else:
            description = suffix

        return dm.DataModelApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name or None,
            version=self.version or "missing",
            description=description,
            views=[],
        )

    def as_identifier(self) -> str:
        return repr(self.as_data_model_id())

    @classmethod
    def _get_description_and_creator(cls, description_raw: str | None) -> tuple[str | None, list[str]]:
        if description_raw and (description_match := re.search(r"Creator: (.+)", description_raw)):
            creator = description_match.group(1).split(", ")
            description = description_raw.replace(description_match.string, "").strip() or None
        elif description_raw:
            creator = ["MISSING"]
            description = description_raw
        else:
            creator = ["MISSING"]
            description = None
        return description, creator

    @classmethod
    def from_data_model(cls, data_model: dm.DataModelApply, has_reference: bool) -> "DMSMetadata":
        description, creator = cls._get_description_and_creator(data_model.description)
        return cls(
            schema_=SchemaCompleteness.complete,
            data_model_type=DataModelType.solution if has_reference else DataModelType.enterprise,
            space=data_model.space,
            name=data_model.name or None,
            description=description,
            external_id=data_model.external_id,
            version=data_model.version,
            creator=creator,
            created=datetime.now(),
            updated=datetime.now(),
        )


class DMSProperty(SheetEntity):
    view: ViewEntity = Field(alias="View")
    view_property: str = Field(alias="View Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    connection: Literal["direct", "edge", "reverse"] | None = Field(None, alias="Connection")
    value_type: DataType | ViewPropertyEntity | ViewEntity | DMSUnknownEntity = Field(alias="Value Type")
    nullable: bool | None = Field(default=None, alias="Nullable")
    is_list: bool | None = Field(default=None, alias="Is List")
    default: str | int | dict | None = Field(None, alias="Default")
    reference: URLEntity | ReferenceEntity | None = Field(default=None, alias="Reference", union_mode="left_to_right")
    container: ContainerEntity | None = Field(None, alias="Container")
    container_property: str | None = Field(None, alias="Container Property")
    index: StrListType | None = Field(None, alias="Index")
    constraint: StrListType | None = Field(None, alias="Constraint")
    class_: ClassEntity = Field(alias="Class (linage)")
    property_: PropertyType = Field(alias="Property (linage)")

    @field_validator("nullable")
    def direct_relation_must_be_nullable(cls, value: Any, info: ValidationInfo) -> None:
        if info.data.get("connection") == "direct" and value is False:
            raise ValueError("Direct relation must be nullable")
        return value

    @field_validator("value_type", mode="after")
    def connections_value_type(
        cls, value: ViewPropertyEntity | ViewEntity | DMSUnknownEntity, info: ValidationInfo
    ) -> DataType | ViewPropertyEntity | ViewEntity | DMSUnknownEntity:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and not isinstance(value, ViewEntity | DMSUnknownEntity):
            raise ValueError(f"Direct relation must have a value type that points to a view, got {value}")
        elif connection == "edge" and not isinstance(value, ViewEntity):
            raise ValueError(f"Edge connection must have a value type that points to a view, got {value}")
        elif connection == "reverse" and not isinstance(value, ViewPropertyEntity | ViewEntity):
            raise ValueError(
                f"Reverse connection must have a value type that points to a view or view property, got {value}"
            )
        return value

    @field_serializer("value_type", when_used="always")
    @staticmethod
    def as_dms_type(value_type: DataType | ViewPropertyEntity | ViewEntity) -> str:
        if isinstance(value_type, DataType):
            return value_type.dms._type
        else:
            return str(value_type)


class DMSContainer(SheetEntity):
    container: ContainerEntity = Field(alias="Container")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    constraint: ContainerEntityList | None = Field(None, alias="Constraint")
    class_: ClassEntity = Field(alias="Class (linage)")

    def as_container(self) -> dm.ContainerApply:
        container_id = self.container.as_id()
        constraints: dict[str, dm.Constraint] = {}
        for constraint in self.constraint or []:
            requires = dm.RequiresConstraint(constraint.as_id())
            constraints[f"{constraint.space}_{constraint.external_id}"] = requires

        return dm.ContainerApply(
            space=container_id.space,
            external_id=container_id.external_id,
            name=self.name or None,
            description=self.description,
            constraints=constraints or None,
            properties={},
        )

    @classmethod
    def from_container(cls, container: dm.ContainerApply) -> "DMSContainer":
        constraints: list[ContainerEntity] = []
        for _, constraint_obj in (container.constraints or {}).items():
            if isinstance(constraint_obj, dm.RequiresConstraint):
                constraints.append(ContainerEntity.from_id(constraint_obj.require))
            # UniquenessConstraint it handled in the properties
        container_entity = ContainerEntity.from_id(container.as_id())
        return cls(
            class_=container_entity.as_class(),
            container=container_entity,
            name=container.name or None,
            description=container.description,
            constraint=constraints or None,
        )


class DMSView(SheetEntity):
    view: ViewEntity = Field(alias="View")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    implements: ViewEntityList | None = Field(None, alias="Implements")
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    filter_: HasDataFilter | NodeTypeFilter | RawFilter | None = Field(None, alias="Filter")
    in_model: bool = Field(True, alias="In Model")
    class_: ClassEntity = Field(alias="Class (linage)")

    def as_view(self) -> dm.ViewApply:
        view_id = self.view.as_id()
        implements = [parent.as_id() for parent in self.implements or []] or None
        if implements is None and isinstance(self.reference, ReferenceEntity):
            # Fallback to the reference if no implements are provided
            parent = self.reference.as_view_id()
            if (parent.space, parent.external_id) != (view_id.space, view_id.external_id):
                implements = [parent]

        return dm.ViewApply(
            space=view_id.space,
            external_id=view_id.external_id,
            version=view_id.version or _DEFAULT_VERSION,
            name=self.name or None,
            description=self.description,
            implements=implements,
            properties={},
        )

    @classmethod
    def from_view(cls, view: dm.ViewApply, in_model: bool) -> "DMSView":
        view_entity = ViewEntity.from_id(view.as_id())
        class_entity = view_entity.as_class(skip_version=True)

        return cls(
            class_=class_entity,
            view=view_entity,
            description=view.description,
            name=view.name,
            implements=[ViewEntity.from_id(parent, _DEFAULT_VERSION) for parent in view.implements] or None,
            in_model=in_model,
        )


class DMSRules(BaseRules):
    metadata: DMSMetadata = Field(alias="Metadata")
    properties: SheetList[DMSProperty] = Field(alias="Properties")
    views: SheetList[DMSView] = Field(alias="Views")
    containers: SheetList[DMSContainer] | None = Field(None, alias="Containers")
    last: "DMSRules | None" = Field(None, alias="Last", description="The previous version of the data model")
    reference: "DMSRules | None" = Field(None, alias="Reference")

    @field_validator("reference")
    def check_reference_of_reference(cls, value: "DMSRules | None", info: ValidationInfo) -> "DMSRules | None":
        if value is None:
            return None
        if value.reference is not None:
            raise ValueError("Reference rules cannot have a reference")
        if value.metadata.data_model_type == DataModelType.solution and (metadata := info.data.get("metadata")):
            warnings.warn(
                issues.dms.SolutionOnTopOfSolutionModelWarning(
                    metadata.as_data_model_id(), value.metadata.as_data_model_id()
                ),
                stacklevel=2,
            )
        return value

    @field_validator("views")
    def matching_version_and_space(cls, value: SheetList[DMSView], info: ValidationInfo) -> SheetList[DMSView]:
        if not (metadata := info.data.get("metadata")):
            return value
        model_version = metadata.version
        if different_version := [view.view.as_id() for view in value if view.view.version != model_version]:
            warnings.warn(issues.dms.ViewModelVersionNotMatchingWarning(different_version, model_version), stacklevel=2)
        if different_space := [view.view.as_id() for view in value if view.view.space != metadata.space]:
            warnings.warn(issues.dms.ViewModelSpaceNotMatchingWarning(different_space, metadata.space), stacklevel=2)
        return value

    @model_validator(mode="after")
    def post_validation(self) -> "DMSRules":
        from ._validation import DMSPostValidation

        issue_list = DMSPostValidation(self).validate()
        if issue_list.warnings:
            issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError([error for error in issue_list if isinstance(error, issues.NeatValidationError)])
        return self

    def dump(
        self,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        as_reference: bool = False,
    ) -> dict[str, Any]:
        from ._serializer import _DMSRulesSerializer

        dumped = self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        space, version = self.metadata.space, self.metadata.version
        serializer = _DMSRulesSerializer(by_alias, space, version)
        clean = serializer.clean(dumped, as_reference)
        last = "Last" if by_alias else "last"
        if last_dump := clean.get(last):
            clean[last] = serializer.clean(last_dump, False)
        reference = "Reference" if by_alias else "reference"
        if self.reference and (ref_dump := clean.get(reference)):
            space, version = self.reference.metadata.space, self.reference.metadata.version
            clean[reference] = _DMSRulesSerializer(by_alias, space, version).clean(ref_dump, True)
        return clean

    def as_schema(self, include_pipeline: bool = False, instance_space: str | None = None) -> DMSSchema:
        from ._exporter import _DMSExporter

        return _DMSExporter(self, include_pipeline, instance_space).to_schema()

    def as_information_architect_rules(self) -> "InformationRules":
        from ._converter import _DMSRulesConverter

        return _DMSRulesConverter(self).as_information_architect_rules()

    def as_domain_expert_rules(self) -> DomainRules:
        from ._converter import _DMSRulesConverter

        return _DMSRulesConverter(self).as_domain_rules()



from pathlib import Path

from rdflib.term import Node

from cognite.neat.legacy.rules.models.rules import Rules

from ._base import BaseExporter


class TripleExporter(BaseExporter[list[tuple[Node, Node, Node]]]):
    """
    Exporter for transformation rules instances sheet to RDF triples
    """

    def _export_to_file(self, filepath: Path) -> None:
        raise NotImplementedError("Export to file not implemented")

    def export(self) -> list[tuple[Node, Node, Node]]:
        return get_instances_as_triples(self.rules)


def get_instances_as_triples(transformation_rules: Rules) -> list[tuple[Node, Node, Node]]:
    """
    Converts transformation rules instances sheet to RDF triples

    Args:
        transformation_rules: An instance of TransformationRules pydantic class

    Returns:
        List of triples provided as tuples

    """
    if transformation_rules.instances:
        return [
            (instance.instance, instance.property_, instance.value)  # type: ignore[misc]
            for instance in transformation_rules.instances
        ]
    return []



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast
from urllib.parse import quote

from cognite.client import CogniteClient
from cognite.client.data_classes import Relationship, RelationshipList
from rdflib import RDF, Literal, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import create_sha256_hash


class RelationshipsExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusions Relationships into Neat.

    Args:
        relationships (Iterable[Asset]): An iterable of relationships.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
    """

    def __init__(
        self,
        relationships: Iterable[Relationship],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.relationships = relationships

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(
            cast(Iterable[Relationship], client.relationships(data_set_external_ids=data_set_external_id)), namespace
        )

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(RelationshipList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extracts an asset with the given asset_id."""
        for relationship in self.relationships:
            yield from self._relationship2triples(relationship, self.namespace)

    @classmethod
    def _relationship2triples(cls, relationship: Relationship, namespace: Namespace) -> list[Triple]:
        """Converts an asset to triples."""

        if relationship.external_id and relationship.source_external_id and relationship.target_external_id:
            # relationships do not have an internal id, so we generate one
            id_ = namespace[f"Relationship_{create_sha256_hash(relationship.external_id)}"]

            # Set rdf type
            triples: list[Triple] = [(id_, RDF.type, namespace["Relationship"])]

            # Set source and target types
            if source_type := relationship.source_type:
                triples.append(
                    (
                        id_,
                        namespace.source_type,
                        namespace[source_type.title()],
                    )
                )

            if target_type := relationship.target_type:
                triples.append(
                    (
                        id_,
                        namespace.target_type,
                        namespace[target_type.title()],
                    )
                )

            # Create attributes

            triples.append((id_, namespace.external_id, Literal(relationship.external_id)))

            triples.append(
                (
                    id_,
                    namespace.source_external_id,
                    Literal(relationship.source_external_id),
                )
            )

            triples.append(
                (
                    id_,
                    namespace.target_external_id,
                    Literal(relationship.target_external_id),
                )
            )

            if relationship.start_time:
                triples.append(
                    (
                        id_,
                        namespace.start_time,
                        Literal(datetime.fromtimestamp(relationship.start_time / 1000, timezone.utc)),
                    )
                )

            if relationship.end_time:
                triples.append(
                    (
                        id_,
                        namespace.end_time,
                        Literal(datetime.fromtimestamp(relationship.end_time / 1000, timezone.utc)),
                    )
                )

            if relationship.created_time:
                triples.append(
                    (
                        id_,
                        namespace.created_time,
                        Literal(datetime.fromtimestamp(relationship.created_time / 1000, timezone.utc)),
                    )
                )

            if relationship.last_updated_time:
                triples.append(
                    (
                        id_,
                        namespace.last_updated_time,
                        Literal(datetime.fromtimestamp(relationship.last_updated_time / 1000, timezone.utc)),
                    )
                )

            if relationship.confidence:
                triples.append(
                    (
                        id_,
                        namespace.confidence,
                        Literal(relationship.confidence),
                    )
                )

            if relationship.labels:
                for label in relationship.labels:
                    # external_id can create ill-formed URIs, so we create websafe URIs
                    # since labels do not have internal ids, we use the external_id as the id
                    triples.append((id_, namespace.label, namespace[f"Label_{quote(label.dump()['externalId'])}"]))

            # Create connection
            if relationship.data_set_id:
                triples.append((id_, namespace.dataset, namespace[f"Dataset_{relationship.data_set_id}"]))

            return triples
        return []



import warnings
from collections.abc import Collection, Hashable, Iterable, Sequence
from pathlib import Path
from typing import Literal, TypeAlias, cast

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from cognite.client.data_classes.data_modeling import (
    ContainerApplyList,
    DataModelApply,
    DataModelApplyList,
    DataModelId,
    SpaceApplyList,
    ViewApplyList,
)
from cognite.client.exceptions import CogniteAPIError

from cognite.neat.rules import issues
from cognite.neat.rules._shared import Rules
from cognite.neat.rules.issues import IssueList
from cognite.neat.rules.models import InformationRules
from cognite.neat.rules.models.dms import DMSRules, DMSSchema, PipelineSchema
from cognite.neat.utils.cdf_loaders import (
    ContainerLoader,
    DataModelingLoader,
    DataModelLoader,
    RawDatabaseLoader,
    RawTableLoader,
    ResourceLoader,
    SpaceLoader,
    TransformationLoader,
    ViewLoader,
)
from cognite.neat.utils.upload import UploadResult

from ._base import CDFExporter

Component: TypeAlias = Literal["all", "spaces", "data_models", "views", "containers", "node_types"]


class DMSExporter(CDFExporter[DMSSchema]):
    """Export rules to Cognite Data Fusion's Data Model Storage (DMS) service.

    Args:
        export_components (frozenset[Literal["all", "spaces", "data_models", "views", "containers"]], optional):
            Which components to export. Defaults to frozenset({"all"}).
        include_space (set[str], optional):
            If set, only export components in the given spaces. Defaults to None which means all spaces.
        existing_handling (Literal["fail", "skip", "update", "force"], optional): How to handle existing components.
            Defaults to "update". See below for details.
        export_pipeline (bool, optional): Whether to export the pipeline. Defaults to False. This means setting
            up transformations, RAW databases and tables to populate the data model.
        instance_space (str, optional): The space to use for the instance. Defaults to None.
        suppress_warnings (bool, optional): Suppress warnings. Defaults to False.

    ... note::

        - "fail": If any component already exists, the export will fail.
        - "skip": If any component already exists, it will be skipped.
        - "update": If any component already exists, it will be updated.
        - "force": If any component already exists, it will be deleted and recreated.

    """

    def __init__(
        self,
        export_components: Component | Collection[Component] = "all",
        include_space: set[str] | None = None,
        existing_handling: Literal["fail", "skip", "update", "force"] = "update",
        export_pipeline: bool = False,
        instance_space: str | None = None,
        suppress_warnings: bool = False,
    ):
        self.export_components = {export_components} if isinstance(export_components, str) else set(export_components)
        self.include_space = include_space
        self.existing_handling = existing_handling
        self.export_pipeline = export_pipeline
        self.instance_space = instance_space
        self.suppress_warnings = suppress_warnings
        self._schema: DMSSchema | None = None

    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        """Export the rules to a file(s).

        If the file is a directory, the components will be exported to separate files, otherwise they will be
        exported to a zip file.

        Args:
            filepath: Directory or zip file path to export to.
            rules:
        """
        if filepath.is_dir():
            self._export_to_directory(filepath, rules)
        else:
            self._export_to_zip_file(filepath, rules)

    def _export_to_directory(self, directory: Path, rules: Rules) -> None:
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_directory(directory, exclude=exclude, new_line=self._new_line, encoding=self._encoding)

    def _export_to_zip_file(self, filepath: Path, rules: Rules) -> None:
        if filepath.suffix not in {".zip"}:
            warnings.warn("File extension is not .zip, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".zip")
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_zip(filepath, exclude=exclude)

    def _create_exclude_set(self):
        if "all" in self.export_components:
            exclude = set()
        else:
            exclude = {"spaces", "data_models", "views", "containers", "node_types"} - self.export_components
        return exclude

    def export(self, rules: Rules) -> DMSSchema:
        if isinstance(rules, DMSRules):
            dms_rules = rules
        elif isinstance(rules, InformationRules):
            dms_rules = rules.as_dms_architect_rules()
        else:
            raise ValueError(f"{type(rules).__name__} cannot be exported to DMS")
        return dms_rules.as_schema(include_pipeline=self.export_pipeline, instance_space=self.instance_space)

    def delete_from_cdf(self, rules: Rules, client: CogniteClient, dry_run: bool = False) -> Iterable[UploadResult]:
        to_export = self._prepare_exporters(rules, client)

        # we need to reverse order in which we are picking up the items to delete
        # as they are sorted in the order of creation and we need to delete them in reverse order
        for items, loader in reversed(to_export):
            item_ids = loader.get_ids(items)
            existing_items = loader.retrieve(item_ids)
            existing_ids = loader.get_ids(existing_items)
            to_delete: list[Hashable] = []
            for item_id in item_ids:
                if (
                    isinstance(loader, DataModelingLoader)
                    and self.include_space is not None
                    and not loader.in_space(item_id, self.include_space)
                ):
                    continue

                if item_id in existing_ids:
                    to_delete.append(item_id)

            deleted: set[Hashable] = set()
            failed_deleted: set[Hashable] = set()
            error_messages: list[str] = []
            if dry_run:
                deleted.update(to_delete)
            elif to_delete:
                try:
                    loader.delete(to_delete)
                except CogniteAPIError as e:
                    failed_deleted.update(loader.get_id(item) for item in e.failed + e.unknown)
                    deleted.update(loader.get_id(item) for item in e.successful)
                    error_messages.append(f"Failed delete: {e.message}")
                else:
                    deleted.update(to_delete)

            yield UploadResult(
                name=loader.resource_name,
                deleted=deleted,
                failed_deleted=failed_deleted,
                error_messages=error_messages,
            )

    def export_to_cdf_iterable(
        self, rules: Rules, client: CogniteClient, dry_run: bool = False
    ) -> Iterable[UploadResult]:
        to_export = self._prepare_exporters(rules, client)

        redeploy_data_model = False
        for items, loader in to_export:
            # The conversion from DMS to GraphQL does not seem to be triggered even if the views
            # are changed. This is a workaround to force the conversion.
            is_redeploying = loader is DataModelingLoader and redeploy_data_model

            to_create, to_delete, to_update, unchanged = self._categorize_items_for_upload(
                loader, items, is_redeploying
            )

            issue_list = IssueList()
            warning_list = self._validate(loader, items)
            issue_list.extend(warning_list)

            created: set[Hashable] = set()
            skipped: set[Hashable] = set()
            changed: set[Hashable] = set()
            failed_created: set[Hashable] = set()
            failed_changed: set[Hashable] = set()
            error_messages: list[str] = []
            if dry_run:
                if self.existing_handling in ["update", "force"]:
                    changed.update(loader.get_id(item) for item in to_update)
                elif self.existing_handling == "skip":
                    skipped.update(loader.get_id(item) for item in to_update)
                elif self.existing_handling == "fail":
                    failed_changed.update(loader.get_id(item) for item in to_update)
                else:
                    raise ValueError(f"Unsupported existing_handling {self.existing_handling}")
            else:
                if to_delete:
                    try:
                        loader.delete(to_delete)
                    except CogniteAPIError as e:
                        error_messages.append(f"Failed delete: {e.message}")

                if isinstance(loader, DataModelingLoader):
                    to_create = loader.sort_by_dependencies(to_create)

                try:
                    loader.create(to_create)
                except CogniteAPIError as e:
                    failed_created.update(loader.get_id(item) for item in e.failed + e.unknown)
                    created.update(loader.get_id(item) for item in e.successful)
                    error_messages.append(e.message)
                else:
                    created.update(loader.get_id(item) for item in to_create)

                if self.existing_handling in ["update", "force"]:
                    try:
                        loader.update(to_update)
                    except CogniteAPIError as e:
                        failed_changed.update(loader.get_id(item) for item in e.failed + e.unknown)
                        changed.update(loader.get_id(item) for item in e.successful)
                        error_messages.append(e.message)
                    else:
                        changed.update(loader.get_id(item) for item in to_update)
                elif self.existing_handling == "skip":
                    skipped.update(loader.get_id(item) for item in to_update)
                elif self.existing_handling == "fail":
                    failed_changed.update(loader.get_id(item) for item in to_update)

            yield UploadResult(
                name=loader.resource_name,
                created=created,
                changed=changed,
                unchanged={loader.get_id(item) for item in unchanged},
                skipped=skipped,
                failed_created=failed_created,
                failed_changed=failed_changed,
                error_messages=error_messages,
                issues=issue_list,
            )

            if loader is ViewLoader and (created or changed):
                redeploy_data_model = True

    def _categorize_items_for_upload(
        self, loader: ResourceLoader, items: Sequence[CogniteResource], is_redeploying
    ) -> tuple[list[CogniteResource], list[CogniteResource], list[CogniteResource], list[CogniteResource]]:
        item_ids = loader.get_ids(items)
        cdf_items = loader.retrieve(item_ids)
        cdf_item_by_id = {loader.get_id(item): item for item in cdf_items}
        to_create, to_update, unchanged, to_delete = [], [], [], []
        for item in items:
            if (
                isinstance(loader, DataModelingLoader)
                and self.include_space is not None
                and not loader.in_space(item, self.include_space)
            ):
                continue

            cdf_item = cdf_item_by_id.get(loader.get_id(item))
            if cdf_item is None:
                to_create.append(item)
            elif is_redeploying:
                to_update.append(item)
                to_delete.append(cdf_item)
            elif loader.are_equal(item, cdf_item):
                unchanged.append(item)
            else:
                to_update.append(item)
        return to_create, to_delete, to_update, unchanged

    def _prepare_exporters(self, rules, client) -> list[tuple[CogniteResourceList, ResourceLoader]]:
        schema = self.export(rules)
        to_export: list[tuple[CogniteResourceList, ResourceLoader]] = []
        if self.export_components.intersection({"all", "spaces"}):
            to_export.append((SpaceApplyList(schema.spaces.values()), SpaceLoader(client)))
        if self.export_components.intersection({"all", "containers"}):
            to_export.append((ContainerApplyList(schema.containers.values()), ContainerLoader(client)))
        if self.export_components.intersection({"all", "views"}):
            to_export.append((ViewApplyList(schema.views.values()), ViewLoader(client, self.existing_handling)))
        if self.export_components.intersection({"all", "data_models"}):
            to_export.append((DataModelApplyList([schema.data_model]), DataModelLoader(client)))
        if isinstance(schema, PipelineSchema):
            to_export.append((schema.databases, RawDatabaseLoader(client)))
            to_export.append((schema.raw_tables, RawTableLoader(client)))
            to_export.append((schema.transformations, TransformationLoader(client)))
        return to_export

    def _validate(self, loader: ResourceLoader, items: CogniteResourceList) -> IssueList:
        issue_list = IssueList()
        if isinstance(loader, DataModelLoader):
            models = cast(list[DataModelApply], items)
            if other_models := self._exist_other_data_models(loader, models):
                warning = issues.dms.OtherDataModelsInSpaceWarning(models[0].space, other_models)
                if not self.suppress_warnings:
                    warnings.warn(warning, stacklevel=2)
                issue_list.append(warning)

        return issue_list

    @classmethod
    def _exist_other_data_models(cls, loader: DataModelLoader, models: list[DataModelApply]) -> list[DataModelId]:
        if not models:
            return []
        space = models[0].space
        external_id = models[0].external_id
        try:
            data_models = loader.client.data_modeling.data_models.list(space=space, limit=25, all_versions=False)
        except CogniteAPIError as e:
            warnings.warn(issues.importing.APIWarning(str(e)), stacklevel=2)
            return []
        else:
            return [
                data_model.as_id()
                for data_model in data_models
                if (data_model.space, data_model.external_id) != (space, external_id)
            ]



"""This module contains the definition of `TransformationRules` pydantic model and all
its sub-models and validators.
"""

from __future__ import annotations

import math
import sys
import types
from abc import abstractmethod
from collections.abc import Callable, Iterator
from functools import wraps
from typing import Annotated, Any, ClassVar, Generic, Literal, TypeAlias, TypeVar

import pandas as pd
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    HttpUrl,
    PlainSerializer,
    constr,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo
from pydantic.main import IncEx

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


METADATA_VALUE_MAX_LENGTH = 5120


def _add_alias(data: dict[str, Any], base_model: type[BaseModel]) -> None:
    for field_name, field_ in base_model.model_fields.items():
        if field_name not in data and field_.alias in data:
            data[field_name] = data[field_.alias]


def replace_nan_floats_with_default(values: dict, model_fields: dict[str, FieldInfo]) -> dict:
    output = {}
    for field_name, value in values.items():
        is_nan_float = isinstance(value, float) and math.isnan(value)
        if not is_nan_float:
            output[field_name] = value
            continue
        if field_name in model_fields:
            output[field_name] = model_fields[field_name].default
        else:
            # field_name may be an alias
            source_name = next((name for name, field in model_fields.items() if field.alias == field_name), None)
            if source_name:
                output[field_name] = model_fields[source_name].default
            else:
                # Just pass it through if it is not an alias.
                output[field_name] = value
    return output


def skip_field_validator(validators_field):
    def decorator(func):
        @wraps(func)
        def wrapper(cls, value, values):
            if isinstance(values, dict):
                to_skip = values.get(validators_field, set())
            else:
                try:
                    to_skip = values.data.get(validators_field, set())
                except Exception:
                    to_skip = set()

            if "all" in to_skip or func.__name__ in to_skip:
                return value
            return func(cls, value, values)

        return wrapper

    return decorator


def skip_model_validator(validators_field):
    def decorator(func):
        @wraps(func)
        def wrapper(self):
            to_skip = getattr(self, validators_field, set())
            if "all" in to_skip or func.__name__ in to_skip:
                return self

            return func(self)

        return wrapper

    return decorator


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


Space: TypeAlias = str
Description: TypeAlias = constr(min_length=1, max_length=1024)  # type: ignore[valid-type]


class SchemaCompleteness(StrEnum):
    complete = "complete"
    partial = "partial"
    extended = "extended"


class ExtensionCategory(StrEnum):
    addition = "addition"
    reshape = "reshape"
    rebuild = "rebuild"


class DataModelType(StrEnum):
    solution = "solution"
    enterprise = "enterprise"


class RoleTypes(StrEnum):
    domain_expert = "domain expert"
    information_architect = "information architect"
    asset_architect = "asset architect"
    dms_architect = "DMS Architect"


class MatchType(StrEnum):
    exact = "exact"
    partial = "partial"


class RuleModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="ignore",
        use_enum_values=True,
    )
    validators_to_skip: set[str] = Field(default_factory=set, exclude=True)

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @classmethod
    def sheets(cls, by_alias: bool = False) -> list[str]:
        """Returns a list of sheet names for the model."""
        return [
            (field.alias or field_name) if by_alias else field_name
            for field_name, field in cls.model_fields.items()
            if field_name != "validators_to_skip"
        ]

    @classmethod
    def headers_by_sheet(cls, by_alias: bool = False) -> dict[str, list[str]]:
        """Returns a list of headers for the model."""
        headers_by_sheet: dict[str, list[str]] = {}
        for field_name, field in cls.model_fields.items():
            if field_name == "validators_to_skip":
                continue
            sheet_name = (field.alias or field_name) if by_alias else field_name
            annotation = field.annotation

            if isinstance(annotation, types.UnionType):
                annotation = annotation.__args__[0]

            try:
                if isinstance(annotation, type) and issubclass(annotation, SheetList):
                    # We know that this is a SheetList, so we can safely access the annotation
                    # which is the concrete type of the SheetEntity.
                    model_fields = annotation.model_fields["data"].annotation.__args__[0].model_fields  # type: ignore[union-attr]
                elif isinstance(annotation, type) and issubclass(annotation, BaseModel):
                    model_fields = annotation.model_fields
                else:
                    model_fields = {}
            except TypeError:
                # Python 3.10 raises TypeError: issubclass() arg 1 must be a class
                # when calling issubclass(annotation, SheetList) with the dict annotation
                model_fields = {}
            headers_by_sheet[sheet_name] = [
                (field.alias or field_name) if by_alias else field_name
                for field_name, field in model_fields.items()
                if field_name != "validators_to_skip"
            ]
        return headers_by_sheet


class URL(BaseModel):
    url: HttpUrl


class BaseMetadata(RuleModel):
    """
    Metadata model for data model
    """

    role: ClassVar[RoleTypes]

    def to_pandas(self) -> pd.Series:
        """Converts Metadata to pandas Series."""
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of Metadata."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @model_serializer(mode="wrap")
    def include_role(self, serializer: Callable) -> dict:
        return {"role": self.role.value, **serializer(self)}

    @abstractmethod
    def as_identifier(self) -> str:
        """Returns a unique identifier for the metadata."""
        raise NotImplementedError()


class BaseRules(RuleModel):
    """
    Rules is a core concept in `neat`. This represents fusion of data model
    definitions and (optionally) the transformation rules used to transform the data/graph
    from the source representation to the target representation defined by the data model.
    The rules are defined in a Excel sheet and then parsed into a `Rules` object. The
    `Rules` object is then used to generate data model and the`RDF` graph made of data
    model instances.

    Args:
        metadata: Data model metadata
        validators_to_skip: List of validators to skip. Defaults to []
    """

    metadata: BaseMetadata

    def dump(
        self,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        as_reference: bool = False,
    ) -> dict[str, Any]:
        """Dump the model to a dictionary.

        This is used in the Exporters to dump rules in the required format.
        """
        return self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )


# An sheet entity is either a class or a property.
class SheetEntity(RuleModel):
    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value


T_Entity = TypeVar("T_Entity", bound=SheetEntity)


class SheetList(BaseModel, Generic[T_Entity]):
    data: list[T_Entity] = Field(default_factory=list)

    @model_validator(mode="before")
    def from_list_format(cls, values: Any) -> Any:
        if isinstance(values, list):
            return {"data": values}
        return values

    def __contains__(self, item: str) -> bool:
        return item in self.data

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[T_Entity]:  # type: ignore[override]
        return iter(self.data)

    def append(self, value: T_Entity) -> None:
        self.data.append(value)

    def extend(self, values: list[T_Entity]) -> None:
        self.data.extend(values)

    def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
        """Converts ResourceDict to pandas DataFrame."""
        df = pd.DataFrame([entity.model_dump() for entity in self.data])
        if drop_na_columns:
            df = df.dropna(axis=1, how="all")
        if include is not None:
            df = df[include]
        return df

    def _repr_html_(self) -> str:
        """Returns HTML representation of ResourceDict."""
        return self.to_pandas(drop_na_columns=True)._repr_html_()  # type: ignore[operator]

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)


ExtensionCategoryType = Annotated[
    ExtensionCategory,
    PlainSerializer(
        lambda v: v.value if isinstance(v, ExtensionCategory) else v,
        return_type=str,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: ExtensionCategory(v) if isinstance(v, str) else v),
]



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import TimeSeries, TimeSeriesList
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import string_to_ideal_type


class TimeSeriesExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusions TimeSeries into Neat.

    Args:
        timeseries (Iterable[TimeSeries]): An iterable of timeseries.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
    """

    def __init__(
        self,
        timeseries: Iterable[TimeSeries],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.timeseries = timeseries

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(
            cast(Iterable[TimeSeries], client.time_series(data_set_external_ids=data_set_external_id)), namespace
        )

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(TimeSeriesList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extract timeseries as triples."""
        for timeseries in self.timeseries:
            yield from self._timeseries2triples(timeseries, self.namespace)

    @classmethod
    def _timeseries2triples(cls, timeseries: TimeSeries, namespace: Namespace) -> list[Triple]:
        id_ = namespace[f"TimeSeries_{timeseries.id}"]

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, namespace.TimeSeries)]

        # Create attributes

        if timeseries.external_id:
            triples.append((id_, namespace.external_id, Literal(timeseries.external_id)))

        if timeseries.name:
            triples.append((id_, namespace.name, Literal(timeseries.name)))

        if timeseries.is_string:
            triples.append((id_, namespace.is_string, Literal(timeseries.is_string)))

        if timeseries.metadata:
            for key, value in timeseries.metadata.items():
                if value:
                    type_aware_value = string_to_ideal_type(value)
                    try:
                        triples.append((id_, namespace[key], URIRef(str(AnyHttpUrl(type_aware_value)))))  # type: ignore
                    except ValidationError:
                        triples.append((id_, namespace[key], Literal(type_aware_value)))

        if timeseries.unit:
            triples.append((id_, namespace.unit, Literal(timeseries.unit)))

        if namespace.is_step:
            triples.append((id_, namespace.is_step, Literal(timeseries.is_step)))

        if timeseries.description:
            triples.append((id_, namespace.description, Literal(timeseries.description)))

        if timeseries.security_categories:
            for category in timeseries.security_categories:
                triples.append((id_, namespace.security_categories, Literal(category)))

        if timeseries.created_time:
            triples.append(
                (
                    id_,
                    namespace.created_time,
                    Literal(datetime.fromtimestamp(timeseries.created_time / 1000, timezone.utc)),
                )
            )

        if timeseries.last_updated_time:
            triples.append(
                (
                    id_,
                    namespace.last_updated_time,
                    Literal(datetime.fromtimestamp(timeseries.last_updated_time / 1000, timezone.utc)),
                )
            )

        if timeseries.legacy_name:
            triples.append((id_, namespace.legacy_name, Literal(timeseries.legacy_name)))

        # Create connections
        if timeseries.unit_external_id:
            # try to create connection to QUDT unit catalog
            try:
                triples.append((id_, namespace.unit_external_id, URIRef(str(AnyHttpUrl(timeseries.unit_external_id)))))
            except ValidationError:
                triples.append((id_, namespace.unit_external_id, Literal(timeseries.unit_external_id)))

        if timeseries.data_set_id:
            triples.append((id_, namespace.dataset, namespace[f"Dataset_{timeseries.data_set_id}"]))

        if timeseries.asset_id:
            triples.append((id_, namespace.asset, namespace[f"Asset_{timeseries.asset_id}"]))

        return triples



import sys
from typing import Any

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

from pydantic import BaseModel, Field, field_validator


class WorkflowState(StrEnum):
    CREATED = "CREATED"
    RUNNING = "RUNNING"
    RUNNING_WAITING = "RUNNING_WAITING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    EXPIRED = "EXPIRED"
    UNKNOWN = "UNKNOWN"


class StepExecutionStatus(StrEnum):
    SUCCESS = "COMPLETED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    STARTED = "STARTED"
    UNKNOWN = "UNKNOWN"
    ABORT_AND_FAIL = "ABORT_AND_FAIL"


class InstanceStartMethod(StrEnum):
    PERSISTENT_INSTANCE_NON_BLOCKING = "persistent_non_blocking"
    PERSISTENT_INSTANCE_BLOCKING = "persistent_blocking"
    EPHEMERAL_INSTANCE = "ephemeral_instance"


class WorkflowStartException(Exception):
    pass


class FlowMessage(BaseModel):
    """A message that can be sent between steps in a workflow.It's the only parameter step takes as input."""

    payload: Any = None  # The payload of the message
    headers: dict[str, str] | None = None  # The headers of the message
    output_text: str | None = None  # The output text of the step that is captured in the execution log
    error_text: str | None = None  # The error text of the step that is captured in the execution log
    next_step_ids: list[str] | None = (
        None  # If set, the workflow will skip default route and go to the next step in the list
    )
    step_execution_status: StepExecutionStatus = StepExecutionStatus.UNKNOWN  # The status of the step execution


class StepType(StrEnum):
    PYSTEP = "pystep"
    STD_STEP = "stdstep"
    START_WORKFLOW_TASK_STEP = "start_workflow_task_step"  # Call a workflow from another workflow
    CONTAINER_TASK_STEP = "container_step"  # TODO: implement
    CLI_TASK_STEP = "cli_step"  # TODO: implement
    HTTP_TRIGGER = "http_trigger"
    TIME_TRIGGER = "time_trigger"
    WAIT_FOR_EVENT = "wait_for_event"  # TODO: implement
    EVENT_TRIGGER = "event_trigger"  # TODO: implement


class UIConfig(BaseModel):
    pos_x: int = 0
    pos_y: int = 0


class WorkflowConfigItem(BaseModel):
    name: str
    value: str | None = None
    label: str | None = None
    type: str | None = None
    required: bool = False
    options: list[str] | None = None
    group: str | None = None


class WorkflowStepDefinition(BaseModel):
    id: str
    label: str | None = None
    stype: str = StepType.PYSTEP
    description: str | None = None
    method: str | None = None
    enabled: bool = True
    system_component_id: str | None = None
    trigger: bool = False
    transition_to: list[str] | None = None
    ui_config: UIConfig = UIConfig()
    params: dict[str, str] = Field(default_factory=dict)  # System parameters
    configs: dict[str, Any] = Field(default_factory=dict)  # Step configurations
    complex_configs: dict[str, Any] = Field(default_factory=dict)  # Complex step configurations
    max_retries: int = 0
    retry_delay: int = 3
    auto_workflow_cleanup: bool = False

    @field_validator("configs", "params", mode="before")
    def none_as_empty_dict(cls, value):
        if value is None:
            return {}
        return value


class WorkflowSystemComponent(BaseModel):
    # Container for steps
    id: str
    label: str
    transition_to: list[str] | None = None
    description: str | None = None
    ui_config: UIConfig = UIConfig()


class WorkflowDefinition(BaseModel):
    """Workflow definition"""

    name: str
    description: str | None = None
    implementation_module: str | None = None
    steps: list[WorkflowStepDefinition] = Field(default_factory=list)
    system_components: list[WorkflowSystemComponent] = Field(default_factory=list)
    configs: list[WorkflowConfigItem] = Field(default_factory=list)

    def get_config_item(self, name: str) -> WorkflowConfigItem | None:
        for config in self.configs:
            if config.name == name:
                return config
        return None

    def upsert_config_item(self, config_item: WorkflowConfigItem) -> None:
        for config in self.configs:
            if config.name == config_item.name:
                config.value = config_item.value
                config.label = config_item.label
                config.type = config_item.type
                config.required = config_item.required
                config.options = config_item.options
                config.group = config_item.group
                return
        self.configs.append(config_item)


class WorkflowStepEvent(BaseModel):
    """Workflow step event represent a single step execution"""

    id: str
    system_component_id: str | None = None
    state: StepExecutionStatus
    elapsed_time: float
    timestamp: str
    error: str | None = None
    output_text: str | None = None
    data: Any = None


class WorkflowFullStateReport(BaseModel):
    """Workflow state report is complete log of workflow execution"""

    workflow_name: str | None = None
    workflow_version: str | None = None
    run_id: str | None = None
    state: WorkflowState
    start_time: int | None = None
    end_time: int | None = None
    elapsed_time: float = 0
    last_error: str | None = None
    execution_log: list[WorkflowStepEvent]
    last_updated_time: int | None = None

    @field_validator("start_time", "end_time", mode="before")
    def float_to_int(cls, value):
        if isinstance(value, float):
            return int(value)
        return value


class WorkflowConfigs(BaseModel):
    """Workflow configs"""

    configs: list[WorkflowConfigItem] = []

    def get_config_item(self, config_name: str) -> WorkflowConfigItem | None:
        return next((item for item in self.configs if item.name == config_name), None)

    def set_config_item(self, config_item: WorkflowConfigItem):
        for item in self.configs:
            if item.name == config_item.name:
                item.value = config_item.value
                return
        self.configs.append(config_item)

    def get_config_group_values_by_name(
        self, group_name: str, remove_group_prefix: bool = True
    ) -> dict[str, str | None]:
        return {
            (item.name.removeprefix(item.group) if remove_group_prefix else item.name): item.value
            for item in self.configs
            if item.group == group_name
        }

    def get_config_item_value(self, config_name: str, default_value=None) -> str | None:
        return config.value if (config := self.get_config_item(config_name)) else default_value



import sys
from collections.abc import Iterable
from typing import ClassVar, cast

from pydantic import BaseModel, ConfigDict, Field
from rdflib import Graph, Literal, Namespace
from rdflib.term import URIRef

from cognite.neat.constants import PREFIXES
from cognite.neat.rules.models._rdfpath import (
    Hop,
    Step,
)
from cognite.neat.utils.utils import remove_namespace, uri_to_short_form

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Triple(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    subject: str | URIRef
    predicate: str | URIRef
    object: str | URIRef | Literal | None = None
    optional: bool = Field(
        description="Indicates whether a triple is optional, used when building SPARQL query",
        default=False,
    )

    @classmethod
    def from_rdflib_triple(cls, triple: tuple[URIRef, URIRef, URIRef | Literal]) -> Self:
        return cls(subject=triple[0], predicate=triple[1], object=triple[2])


def generate_prefix_header(prefixes: dict[str, Namespace] = PREFIXES) -> str:
    """Generate prefix header which is added to SPARQL query and allows for shorten query statements

    Parameters
    ----------
    prefixes : dict
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    str
        Prefix header
    """
    return "".join(f"PREFIX {key}:<{value}>\n" for key, value in prefixes.items())


def get_predicate_id(
    graph: Graph, subject_type_id: str, object_type_id: str, prefixes: dict[str, Namespace] = PREFIXES
) -> URIRef:
    """Returns predicate (aka property) URI (i.e., ID) that connects subject and object

    Parameters
    ----------
    graph : Graph
        Data model graph or data model instance (aka knowledge graph)
    subject_type_id : str
        ID of subject type (aka subject class)
    object_type_id : str
        ID of object type (aka object class)
    prefixes : dict, optional
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    URIRef
        ID of predicate (aka property) connecting subject and object
    """
    query = """

        SELECT ?predicateTypeID
        WHERE {
            ?subjectInstanceID a subjectTypeID .
            ?objectInstanceID a objectTypeID .
            ?subjectInstanceID ?predicateTypeID ?objectInstanceID .
        } LIMIT 1"""

    query = query.replace("insertPrefixes", generate_prefix_header(prefixes))
    final_query = query.replace("subjectTypeID", subject_type_id).replace("objectTypeID", object_type_id)
    res = list(cast(tuple, graph.query(final_query)))

    if len(res) != 1:
        raise ValueError("Subject and Object must have exactly 1 relation!")

    return res[0][0]


def hop2property_path(graph: Graph, hop: Hop, prefixes: dict[str, Namespace]) -> str:
    """Converts hop to property path string

    Parameters
    ----------
    graph : Graph
        Graph containing instances of classes
    hop : Hop
        Hop to convert
    prefixes : dict[str, Namespace]
        Dictionary of prefixes to use for compression and predicate querying

    Returns
    -------
    str
        Property path string for hop traversal (e.g. ^rdf:type/rdfs:subClassOf)
    """

    # setting previous step to origin, as we are starting from there
    previous_step = Step(class_=hop.class_, direction="origin")

    # add triples for all steps until destination
    property_path = ""
    for current_step in hop.traversal:
        sub_entity, obj_entity = (
            (current_step, previous_step) if current_step.direction == "source" else (previous_step, current_step)
        )

        predicate_raw = get_predicate_id(graph, sub_entity.class_.id, obj_entity.class_.id, prefixes)

        predicate = uri_to_short_form(predicate_raw, prefixes)

        predicate = f"^{predicate}" if current_step.direction == "source" else predicate
        property_path += f"{predicate}/"

        previous_step = current_step

    if previous_step.property:
        return property_path + previous_step.property.id
    else:
        # removing "/" at the end of property path if there is no property at the end
        return property_path[:-1]


def triples2dictionary(triples: Iterable[tuple[URIRef, URIRef, str | URIRef]]) -> dict[URIRef, dict[str, list[str]]]:
    """Converts list of triples to dictionary"""
    dictionary: dict[URIRef, dict[str, list[str]]] = {}
    for triple in triples:
        id_: str
        property_: str
        value: str
        uri: URIRef

        id_, property_, value = remove_namespace(*triple)  # type: ignore[misc]
        uri = triple[0]

        if uri not in dictionary:
            dictionary[uri] = {"external_id": [id_]}

        if property_ not in dictionary[uri]:
            dictionary[uri][property_] = [value]
        else:
            dictionary[uri][property_].append(value)
    return dictionary



from datetime import datetime, timezone
from typing import Any, Literal

import pandas as pd

from cognite.neat.legacy.rules.models.tables import Tables

from ._base import BaseImporter


class ArbitraryDictImporter(BaseImporter):
    """
    Importer for an arbitrary dictionary.

    This importer infers the data model from the dictionary based on the shape of the data.

    Args:
        data: dictionary containing Rules definitions.
        relationship_direction: Direction of relationships, either "parent-to-child" or "child-to-parent". Dictionaries
            are nested with children nested inside parents. This option determines whether the resulting rules
            will have an edge from parents to children or from children to parents.
    """

    def __init__(
        self,
        data: dict[str, Any],
        relationship_direction: Literal["parent-to-child", "child-to-parent"] = "parent-to-child",
    ):
        self.data = data
        self.relationship_direction = relationship_direction

    def to_tables(self) -> dict[str, pd.DataFrame]:
        metadata = pd.Series(
            dict(
                title="OpenAPI to DM transformation rules",
                description="OpenAPI to DM transformation rules",
                version="0.1",
                creator="Cognite",
                created=datetime.now(timezone.utc).replace(microsecond=0).isoformat(),
                namespace="http://purl.org/cognite/neat#",
                prefix="neat",
                data_model_name="OpenAPI",
                cdf_space_name="OpenAPI",
            )
        ).reset_index()
        finder = _TripleFinder(self.relationship_direction)
        finder.find_triples(self.data)

        return {
            Tables.metadata: metadata,
            Tables.classes: pd.DataFrame(finder.classes).T,
            Tables.properties: pd.DataFrame(finder.properties).T,
        }


class _TripleFinder:
    def __init__(self, relationship_direction: Literal["parent-to-child", "child-to-parent"]) -> None:
        self.classes: dict[str, dict[str, Any]] = {}
        self.properties: dict[str, dict[str, Any]] = {}
        self.relationship_direction = relationship_direction

    def find_triples(self, data: dict[str, Any]) -> None:
        self._convert_dict_to_classes_and_props(data)

    def _convert_dict_to_classes_and_props(
        self, data: dict, parent_property_name: str | None = None, grand_parent_property_name: str | None = None
    ) -> None:
        if isinstance(data, dict) and len(data) == 0:
            return
        elif isinstance(data, dict) and parent_property_name is None:
            for key, value in data.items():
                self._convert_dict_to_classes_and_props(value, key)
        elif isinstance(data, dict):
            self.add_class(parent_property_name, "missing", grand_parent_property_name, is_list=False)
            for key, value in data.items():
                self._convert_dict_to_classes_and_props(value, key, parent_property_name)
        elif isinstance(data, list):
            if parent_property_name is not None and grand_parent_property_name is not None:
                data_type = self._get_list_type(data, parent_property_name)
                self.add_property(grand_parent_property_name, parent_property_name, data_type, "missing", is_list=True)
            for item in data:
                self._convert_dict_to_classes_and_props(item, parent_property_name, grand_parent_property_name)
        elif isinstance(data, bool | int | float | str) and parent_property_name is not None:
            data_type = self._get_primitive_data_type(data)
            self.add_property(grand_parent_property_name, parent_property_name, data_type, "missing")
        else:
            raise ValueError(f"Unknown type {type(data)}")

    def add_class(
        self, class_name: str, description: str = "", parent_class_name: str | None = None, is_list: bool = False
    ):
        if class_name in self.classes:
            return
        class_ = {"Class": class_name, "description": description}
        if parent_class_name:
            if self.relationship_direction == "child-to-parent":
                self.add_property(class_name, "parent", parent_class_name, "missing", is_list=False)
            elif self.relationship_direction == "parent-to-child":
                self.add_property(parent_class_name, class_name, class_name, "missing", is_list)
            else:
                raise ValueError(f"Unknown relationship direction {self.relationship_direction}")
        self.classes[class_name] = class_

    def add_property(
        self,
        class_name: str,
        property_name: str,
        property_type: str,
        description: str = "missing",
        is_list: bool = False,
    ):
        if class_name + property_name in self.properties:
            return
        prop = dict(
            class_id=class_name,
            property_id=property_name,
            property_name=property_name,
            property_type="ObjectProperty",
            description=description,
            expected_value_type=property_type,
            max_count=1 if not is_list else None,
            cdf_resource_type="Asset",
            resource_type_property="Asset",
            rule_type="rdfpath",
            rule=f"neat:{class_name}(neat:{property_name})",
            label="linked to",
        )
        self.properties[class_name + property_name] = prop

    @staticmethod
    def _get_primitive_data_type(data: Any, errors: Literal["raise", "empty"] = "raise") -> str:
        data_type = type(data)
        if data_type is bool:
            return "boolean"
        elif data_type is int:
            return "integer"
        elif data_type is float:
            return "float"
        elif data_type is str and not pd.isna(pd.to_datetime(data, errors="coerce")):
            return "dateTime"
        elif data_type is str:
            return "string"

        if errors == "empty":
            return ""
        else:
            raise ValueError(f"Unknown primitive type {data_type}")

    @classmethod
    def _get_list_type(cls, data: list[Any], class_name: str) -> str:
        if isinstance(data[0], dict):
            return class_name

        data_types = {cls._get_primitive_data_type(item, "empty") for item in data}
        if "" in data_types or len(data_types) > 1:
            # Fallback to string
            return "string"
        return data_types.pop()



import logging
import uuid
import warnings
from pathlib import Path
from typing import cast

import numpy as np
import pandas as pd
from openpyxl import Workbook, load_workbook
from openpyxl.cell import Cell
from openpyxl.styles import Alignment, Border, Font, NamedStyle, PatternFill, Side
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat.legacy.graph import exceptions
from cognite.neat.legacy.graph.exceptions import NamespaceRequired
from cognite.neat.legacy.graph.models import Triple
from cognite.neat.legacy.rules.analysis import get_defined_classes, to_class_property_pairs
from cognite.neat.legacy.rules.exporters._rules2rules import to_dms_name
from cognite.neat.legacy.rules.models.rules import Rules

from ._base import BaseExtractor


class GraphCapturingSheet(BaseExtractor):
    """
    Graph capturing sheet class that provides methods for creating a graph capturing sheet and extracting RDF triples.

    Args:
        rules: Transformation rules which holds data model that is used to validate
                the graph capturing sheet and extract data model instances from it (i.e. RDF triples)
        filepath: File path to save the sheet to. Defaults to None.
        separator: Multi value separator at cell level. Defaults to ",".
        namespace: Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to None, meaning namespace of rules will be used.
        store_graph_capturing_sheet: Whether to store the graph capturing sheet in the object. Will be stored in the
                                     `sheet` attribute. Defaults to False.
        use_source_ids : Whether to use source ids for properties and classes stored in Source column if they exist.
                         Defaults to False, meaning that the source ids will be ignored.

    """

    def __init__(
        self,
        rules: Rules,
        filepath: Path | str | None = None,
        separator: str = ",",
        namespace: str | None = None,
        store_graph_capturing_sheet: bool = False,
        use_source_ids: bool = False,
    ):
        self.rules = rules
        self.filepath = Path(filepath) if isinstance(filepath, str | Path) else None
        self.separator = separator
        self.namespace = namespace
        self.store_graph_capturing_sheet = store_graph_capturing_sheet
        self.use_source_ids = use_source_ids
        self.sheet: dict[str, pd.DataFrame] = {}

    def create_template(self, filepath: Path | None = None, overwrite: bool = False) -> None:
        """
        Creates a graph capturing sheet template based on the transformation rules.

        Args:
            filepath: File path to save the sheet to. Defaults to None.
            overwrite: Overwrite existing file. Defaults to False.
        """
        if filepath is None:
            filepath = self.filepath
        if filepath is None:
            raise ValueError("File path to the graph capturing sheet is not provided!")
        if filepath.exists() and not overwrite:
            raise FileExistsError(f"File {filepath} already exists! Set overwrite to True to overwrite it!")
        rules2graph_capturing_sheet(self.rules, filepath)

    def extract(self) -> list[Triple]:
        """
        Extracts RDF triples from the graph capturing sheet.

        Returns:
            List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
        """
        if self.filepath is None:
            raise ValueError("File path to the graph capturing sheet is not provided!")
        graph_capturing_sheet = read_graph_excel_file_to_table_by_name(self.filepath)
        if self.store_graph_capturing_sheet:
            self.sheet = graph_capturing_sheet

        print(self.namespace)

        return sheet2triples(graph_capturing_sheet, self.rules, self.separator, self.namespace, self.use_source_ids)


def extract_graph_from_sheet(
    filepath: Path, transformation_rule: Rules, separator: str = ",", namespace: str | None = None
) -> list[Triple]:
    """Converts a graph capturing sheet to RDF triples that define data model instances

    Args:
        filepath : Path to the graph capturing sheet
        transformation_rule : Transformation rules which holds data model that is used to validate
                              the graph capturing sheet and extract data model instances from it (i.e. RDF triples)
        separator : Multi value separator at cell level. Defaults to ",".
        namespace : Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to None, meaning namespace of rules will be used.

    Returns:
        List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
    """

    graph_capturing_sheet = read_graph_excel_file_to_table_by_name(filepath)

    return sheet2triples(graph_capturing_sheet, transformation_rule, separator, namespace)


def sheet2triples(
    graph_capturing_sheet: dict[str, pd.DataFrame],
    rules: Rules,
    separator: str = ",",
    namespace: str | None = None,
    use_source_ids: bool = False,
) -> list[Triple]:
    """Converts a graph capturing sheet represented as dictionary of dataframes to rdf triples

    Args:
        graph_capturing_sheet : Graph capturing sheet provided as dictionary of dataframes
        transformation_rule : Transformation rules which holds data model that is used to validate
                             the graph capturing sheet and extract data model instances from it (i.e. RDF triples)
        separator : Multi value separator at cell level. Defaults to ",".
        namespace : Optional custom namespace to use for extracted triples that define
                    data model instances. Defaults to None, meaning namespace of rules will be used.
        use_source_ids : Whether to use source ids for properties and classes stored in Source column if they exist.
                         Defaults to False, meaning that the source ids will be ignored.

    Returns:
        List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
    """

    # Validation that everything is in order before proceeding
    validate_if_graph_capturing_sheet_empty(graph_capturing_sheet)
    validate_rules_graph_pair(graph_capturing_sheet, rules)

    # get class property pairs
    class_property_pairs = to_class_property_pairs(rules)

    # namespace selection
    if namespace is None and rules.metadata.namespace is not None:
        instance_namespace = rules.metadata.namespace
    elif namespace:
        instance_namespace = Namespace(namespace)
    else:
        raise NamespaceRequired("Extract instances from sheet")

    if rules.metadata.namespace is not None:
        model_namespace = Namespace(rules.metadata.namespace)
    else:
        raise NamespaceRequired("Extract instances from sheet")

    # Now create empty graph
    triples: list[Triple] = []

    # Add triples from the capturing sheet to the graph by iterating over the capturing sheet
    # iterate over sheets
    for sheet_name, df in graph_capturing_sheet.items():
        # iterate over sheet rows

        class_uri = (
            URIRef(str(rules.classes[sheet_name].source))
            if use_source_ids and rules.classes[sheet_name].source
            else model_namespace[sheet_name]
        )

        for _, row in df.iterrows():
            if row.identifier is None:
                msg = f"Missing identifier in sheet {sheet_name} at row {row.name}! Skipping..."
                logging.warning(msg)
                warnings.warn(msg, stacklevel=2)
                continue

            # iterate over sheet rows properties
            for property_name, value in row.to_dict().items():
                # Setting RDF type of the instance
                if property_name == "identifier":
                    triples.append((instance_namespace[row.identifier], RDF.type, class_uri))
                    continue
                elif not value:
                    continue

                property_uri = (
                    URIRef(str(class_property_pairs[sheet_name][property_name].source))
                    if use_source_ids and class_property_pairs[sheet_name][property_name].source
                    else model_namespace[property_name]
                )

                property_ = class_property_pairs[sheet_name][property_name]

                is_one_to_many = separator and (
                    (property_.max_count and property_.max_count > 1) or not property_.max_count
                )

                values = value.split(separator) if is_one_to_many else [value]

                # Adding object properties
                if property_.property_type == "ObjectProperty":
                    triples.extend(
                        (
                            instance_namespace[row.identifier],
                            property_uri,
                            instance_namespace[v.strip()],
                        )
                        for v in values
                    )
                # Adding data properties
                elif property_.property_type == "DatatypeProperty":
                    for v in values:
                        try:
                            literal_value = v.strip()
                        except AttributeError:
                            literal_value = v

                        triples.append(
                            (
                                instance_namespace[row.identifier],
                                property_uri,
                                Literal(literal_value, datatype=XSD[property_.expected_value_type.suffix]),
                            )
                        )

                else:
                    raise exceptions.UnsupportedPropertyType(property_.property_type)
    return triples


def validate_if_graph_capturing_sheet_empty(graph_capturing_sheet: dict[str, pd.DataFrame]):
    """Validate if the graph capturing sheet is empty

    Args:
        graph_capturing_sheet : Graph capturing sheet
    """
    if all(df.empty for df in graph_capturing_sheet.values()):
        msg = "Graph capturing sheet is empty! Aborting!"
        logging.error(msg)
        raise ValueError(msg)


def validate_rules_graph_pair(graph_capturing_sheet: dict[str, pd.DataFrame], transformation_rule: Rules):
    """Validate if the graph capturing sheet is based on the transformation rules

    Args:
        graph_capturing_sheet : Graph capturing sheet
        transformation_rule : Transformation rules
    """
    intersection = set(graph_capturing_sheet.keys()).intersection(set(get_defined_classes(transformation_rule)))

    if not intersection:
        msg = "Graph capturing sheet is not based on transformation rules! Aborting!"
        logging.error(msg)
        raise ValueError(msg)

    elif len(intersection) == len(graph_capturing_sheet.keys()):
        logging.info("All classes in the graph capturing sheet are defined in the transformation rules!")

    elif len(intersection) < len(graph_capturing_sheet.keys()):
        msg = "Graph capturing sheet contains classes that are not defined in the transformation rules! Proceeding..."
        logging.warning(msg)
        warnings.warn(msg, stacklevel=2)

    elif len(intersection) < len(get_defined_classes(transformation_rule)):
        msg = "Transformation rules contain classes that are not present in the graph capturing sheet! Proceeding..."
        logging.warning(msg)
        warnings.warn(msg, stacklevel=2)


def read_graph_excel_file_to_table_by_name(filepath: Path) -> dict[str, pd.DataFrame]:
    workbook: Workbook = load_workbook(filepath)

    parsed_sheets = {
        sheetname: pd.read_excel(filepath, sheet_name=sheetname, header=0) for sheetname in workbook.sheetnames
    }

    for sheetname, df in parsed_sheets.items():
        if "identifier" in df.columns:
            parsed_sheets[sheetname] = df.drop(df[df.identifier == 0].index)
            parsed_sheets[sheetname] = df.replace({np.nan: None})
        else:
            logging.error(f"Sheet {sheetname} does not have an identifier column")
            raise ValueError(f"Sheet {sheetname} does not have an identifier column")

    return parsed_sheets


def rules2graph_capturing_sheet(
    rules: Rules,
    file_path: Path,
    no_rows: int = 1000,
    auto_identifier_type: str = "index-based",
    add_drop_down_list: bool = True,
):
    """
    Converts a TransformationRules object to a graph capturing sheet

    Args:
        rules: The TransformationRules object to convert to the graph capturing sheet
        file_path: File path to save the sheet to
        no_rows: Number of rows for processing, by default 1000
        auto_identifier_type: Type of automatic identifier, by default "index" based, alternative is "uuid" based
        add_drop_down_list: Add drop down selection for columns that contain linking properties, by default True

    !!! note "no_rows parameter"
        no_rows should be set to the maximum expected number of instances of any of the classes.
        By default, it is set to 1000, increase it accordingly if you have more instances.

    """

    workbook = Workbook()
    # Remove default sheet named "Sheet"
    workbook.remove(workbook["Sheet"])

    for class_, properties in to_class_property_pairs(rules).items():
        workbook.create_sheet(title=class_)

        # Add header rows
        cast(Worksheet, workbook[class_]).append(["identifier", *list(properties.keys())])

        if auto_identifier_type and auto_identifier_type == "index-based":  # default, easy to read
            logging.debug(f"Configuring index-based automatic identifiers for sheet {class_}")
            _add_index_identifiers(workbook, class_, no_rows)
        elif auto_identifier_type and auto_identifier_type == "uuid-based":
            logging.debug(f"Configuring UUID-based automatic identifiers for sheet {class_}")
            _add_uuid_identifiers(workbook, class_, no_rows)
        else:
            logging.debug(f"No automatic identifier set for sheet {class_}!")

        for i, property_ in enumerate(properties.values()):
            if property_.property_type == "ObjectProperty" and add_drop_down_list:
                _add_drop_down_list(
                    workbook, class_, get_column_letter(i + 2), no_rows, property_.expected_value_type.suffix, "A"
                )

    _adjust_column_width(workbook)
    _set_header_style(workbook)

    logging.info(f"Graph capturing sheet generated and stored at {file_path}!")
    workbook.save(file_path)
    workbook.close()


def _add_index_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds index-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        prefix = to_dms_name(sheet, "class", True)
        workbook[sheet][f"A{i+2}"] = f'=IF(ISBLANK(B{i+2}), "","{prefix}-{i+1}")'  # type: ignore[index]


def _add_uuid_identifiers(workbook: Workbook, sheet: str, no_rows: int):
    """Adds UUID-based auto identifier to a sheet identifier column"""
    for i in range(no_rows):
        prefix = to_dms_name(sheet, "class", True)
        workbook[sheet][f"A{i+2}"] = f'=IF(ISBLANK(B{i+2}), "","{prefix}-{uuid.uuid4()}")'  # type: ignore[index]


def _add_drop_down_list(workbook: Workbook, sheet: str, column: str, no_rows: int, value_sheet: str, value_column: str):
    """Adds a drop down list to a column"""
    drop_down_list = DataValidation(type="list", formula1=f"={value_sheet}!{value_column}$2:{value_column}${no_rows}")

    cast(Worksheet, workbook[sheet]).add_data_validation(drop_down_list)

    for i in range(no_rows):
        drop_down_list.add(workbook[sheet][f"{column}{i+2}"])  # type: ignore[index, misc]


def _adjust_column_width(workbook: Workbook):
    """Adjusts the column width based on the content"""
    for sheet in workbook.sheetnames:
        for cell_tuple in cast(Worksheet, workbook[sheet]).columns:
            # Wrong type annotation in openpyxl
            cell = cast(Cell, cell_tuple[0])  # type: ignore[index]
            if cell.value:
                adjusted_width = (len(str(cell.value)) + 5) * 1.2
                cast(Worksheet, workbook[sheet]).column_dimensions[cell.column_letter].width = adjusted_width


def _set_header_style(workbook: Workbook):
    """Sets the header style for all sheets in the workbook"""
    style = NamedStyle(name="header style")
    style.font = Font(bold=True, size=16)
    side = Side(style="thin", color="000000")
    style.border = Border(left=side, right=side, top=side, bottom=side)
    workbook.add_named_style(style)

    for sheet in workbook.sheetnames:
        for cell_tuple in cast(Worksheet, workbook[sheet]).columns:
            # Wrong type annotation in openpyxl
            cell = cast(Cell, cell_tuple[0])  # type: ignore[index]
            worksheet = cast(Worksheet, workbook[sheet])
            worksheet[f"{cell.column_letter}1"].style = style
            if f"{cell.column_letter}1" == "A1":
                worksheet[f"{cell.column_letter}1"].fill = PatternFill("solid", start_color="2FB5F2")
            else:
                worksheet[f"{cell.column_letter}1"].fill = PatternFill("solid", start_color="FFB202")
            worksheet[f"{cell.column_letter}1"].alignment = Alignment(horizontal="center", vertical="center")



from collections.abc import Sequence
from dataclasses import dataclass, fields

from cognite.client import CogniteClient
from cognite.client.data_classes import AssetWrite

from cognite.neat.graph._tracking.base import Tracker
from cognite.neat.graph._tracking.log import LogTracker
from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.issues import NeatIssue, NeatIssueList
from cognite.neat.rules.models import AssetRules

from ._base import CDFLoader


@dataclass(frozen=True)
class AssetLoaderMetadataKeys:
    """Class holding mapping between NEAT metadata key names and their desired names
    in CDF Asset metadata

    Args:
        start_time: Start time key name
        end_time: End time key name
        update_time: Update time key name
        resurrection_time: Resurrection time key name
        identifier: Identifier key name
        active: Active key name
        type: Type key name
    """

    start_time: str = "start_time"
    end_time: str = "end_time"
    update_time: str = "update_time"
    resurrection_time: str = "resurrection_time"
    identifier: str = "identifier"
    active: str = "active"
    type: str = "type"

    def as_aliases(self) -> dict[str, str]:
        return {str(field.default): getattr(self, field.name) for field in fields(self)}


class AssetLoader(CDFLoader[AssetWrite]):
    def __init__(
        self,
        rules: AssetRules,
        graph_store: NeatGraphStore,
        data_set_id: int,
        use_orphanage: bool = False,
        use_labels: bool = False,
        asset_external_id_prefix: str | None = None,
        metadata_keys: AssetLoaderMetadataKeys | None = None,
        create_issues: Sequence[NeatIssue] | None = None,
        tracker: type[Tracker] | None = None,
    ):
        super().__init__(graph_store)

        self.rules = rules
        self.data_set_id = data_set_id
        self.use_labels = use_labels
        self.use_orphanage = use_orphanage

        self.orphanage_external_id = (
            f"{asset_external_id_prefix or ''}orphanage-{data_set_id}" if use_orphanage else None
        )

        self.asset_external_id_prefix = asset_external_id_prefix
        self.metadata_keys = metadata_keys or AssetLoaderMetadataKeys()

        self._issues = NeatIssueList[NeatIssue](create_issues or [])
        self._tracker: type[Tracker] = tracker or LogTracker

    @classmethod
    def from_rules(
        cls,
        rules: AssetRules,
        graph_store: NeatGraphStore,
        data_set_id: int,
        use_orphanage: bool = False,
        use_labels: bool = False,
        asset_external_id_prefix: str | None = None,
        metadata_keys: AssetLoaderMetadataKeys | None = None,
    ) -> "AssetLoader":
        issues: list[NeatIssue] = []

        return cls(
            rules, graph_store, data_set_id, use_orphanage, use_labels, asset_external_id_prefix, metadata_keys, issues
        )

    def _create_validation_classes(self) -> None:
        # need to get back class-property pairs where are definition of
        # asset implementations, extend InformationRulesAnalysis make it generic

        # by default if there is not explicitly stated external_id
        # use rdf:type and drop the prefix

        # based on those create pydantic model AssetDefinition
        # which will have .to_asset_write()

        raise NotImplementedError("Not implemented yet, this is placeholder")

    def categorize_assets(self, client: CogniteClient) -> None:
        """Categorize assets to those to be created, updated, decommissioned, or resurrected"""

        raise NotImplementedError("Not implemented yet, this is placeholder")

    def load_to_cdf(self, client: CogniteClient, dry_run: bool = False) -> Sequence[AssetWrite]:
        # generate assets
        # check for circular asset hierarchy
        # check for orphaned assets
        # batch upsert of assets to CDF (otherwise we will hit the API rate limit)

        raise NotImplementedError("Not implemented yet, this is placeholder")

    @classmethod
    def _check_for_circular_asset_hierarchy(cls, assets: list[AssetWrite]) -> None:
        """Check for circular references in the asset rules"""
        raise NotImplementedError("Not implemented yet, this is placeholder")

    @classmethod
    def _check_for_orphaned_assets(cls, assets: list[AssetWrite]) -> None:
        """Check for circular references in the asset rules"""
        raise NotImplementedError("Not implemented yet, this is placeholder")



import logging


class EndpointFilter(logging.Filter):
    """Filter class to exclude specific endpoints from log entries."""

    def __init__(self, excluded_endpoints: list[str]) -> None:
        """
        Initialize the EndpointFilter class.

        Args:
            excluded_endpoints: A list of endpoints to be excluded from log entries.
        """
        self.excluded_endpoints = excluded_endpoints

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Filter out log entries for excluded endpoints.

        Args:
            record: The log record to be filtered.

        Returns:
            bool: True if the log entry should be included, False otherwise.
        """
        return all(endpoint not in record.getMessage() for endpoint in self.excluded_endpoints)



# IO, CDF, Workflows


import logging
import shutil

from fastapi import APIRouter, UploadFile

from cognite.neat.app.api.configuration import NEAT_APP
from cognite.neat.config import Config
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.utils import get_file_hash

router = APIRouter()


@router.get("/api/cdf/neat-resources")
def get_neat_resources(resource_type: str | None = None):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    if resource_type is None:
        return {"error": "Resource type is not specified"}
    result = NEAT_APP.cdf_store.get_list_of_resources_from_cdf(resource_type=resource_type)
    logging.debug(f"Got {len(result)} resources")
    return {"result": result}


@router.post("/api/cdf/init-neat-resources")
def init_neat_cdf_resources(resource_type: str | None = None):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    NEAT_APP.cdf_store.init_cdf_resources(resource_type=resource_type)
    return {"result": "ok"}


@router.post("/api/file/upload/{workflow_name}/{file_type}/{step_id}/{action}")
async def file_upload_handler(
    files: list[UploadFile], workflow_name: str, file_type: str, step_id: str, action: str
) -> dict[str, str]:
    if NEAT_APP.cdf_store is None or NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    if NEAT_APP.workflow_manager.data_store_path is None:
        return {"error": "Workflow Manager is not initialized"}

    # get directory path
    upload_dir = NEAT_APP.config.rules_store_path
    file_name = ""
    file_version = ""
    if file_type == "file_from_editor":
        upload_dir = NEAT_APP.workflow_manager.config.workflows_store_path / workflow_name
    elif file_type == "workflow":
        upload_dir = NEAT_APP.workflow_manager.config.workflows_store_path
    elif file_type == "staging":
        upload_dir = NEAT_APP.workflow_manager.config.staging_path
    elif file_type == "source_graph":
        upload_dir = NEAT_APP.workflow_manager.config.source_graph_path

    for file in files:
        logging.info(
            f"Uploading file : {file.filename} , workflow : {workflow_name} , step_id {step_id} , action : {action}"
        )
        # save file to disk
        if file.filename:
            if file_type == "global_config":
                full_path = NEAT_APP.config.data_store_path / "config.yaml"
            else:
                full_path = upload_dir / file.filename
            with full_path.open("wb") as buffer:
                shutil.copyfileobj(file.file, buffer)
            file_name = file.filename
            file_version = get_file_hash(full_path)
        break  # only one file is supported for now

    if "update_config" in action and file_type == "rules":
        logging.info("Automatically updating workflow config")
        workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
        if workflow is None:
            return {"error": f"Workflow {workflow_name} not found"}
        workflow_definition = workflow.get_workflow_definition()

        for step in workflow_definition.steps:
            if step.method == "ImportExcelToRules":
                step.configs["file_name"] = file_name
                step.configs["version"] = ""

        NEAT_APP.workflow_manager.save_workflow_to_storage(workflow_name)

    if "start_workflow" in action and file_type == "rules" or file_type == "staging":
        logging.info("Starting workflow after file upload")
        workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
        if workflow is None:
            return {"error": f"Workflow {workflow_name} not found"}
        flow_msg = FlowMessage(
            payload={"file_name": file_name, "hash": file_version, "full_path": full_path, "file_type": file_type}
        )
        start_step_id = None if step_id == "none" else step_id

        workflow.start(sync=False, flow_message=flow_msg, start_step_id=start_step_id)

    if action == "install" and file_type == "workflow":
        logging.info("Installing workflow after file upload")
        NEAT_APP.cdf_store.extract_workflow_package(file_name)

    if file_type == "global_config":
        logging.info("Updating global config and restarting NeatApp")
        config = Config.from_yaml(full_path)
        config.data_store_path = NEAT_APP.config.data_store_path
        NEAT_APP.stop()
        NEAT_APP.start(config=config)
        logging.info("NeatApp restarted")

    return {"file_name": file_name, "hash": file_version}
